import sys
import warnings
from functools import wraps
import typing
from typing import Dict, Any, Tuple, Callable, Optional, Union, TypeVar
from copy import deepcopy
import pyterrier as pt


_started = False
_configs = {}


# ----------------------------------------------------------
# Decorators
# ----------------------------------------------------------
# These functions wrap functions/classes to enforce certain
# behavior regarding Java. For instance @pt.java.required
# automatically starts java before it's invoked (if it's not
# already started).
# ----------------------------------------------------------

T = TypeVar("T", bound=Callable[..., Any])

@pt.utils.pre_invocation_decorator
def required(fn: T) -> None:
    """
    Requires the Java Virtual Machine to be started. If the JVM has not yet been started, it runs pt.java.init().

    Can be used as a function/class @decorator. When used as a class decorator, it
    is applied to all methods defined by the class.
    """
    if not _started:
        trigger = fn.__qualname__ if hasattr(fn, '__qualname__') else fn.__name__
        _init(trigger=trigger)

def required_raise(fn: T) -> T:
    """
    Similar to `pt.java.required`, but raises an error if called before pt.java.init().
    """
    @wraps(fn)
    def _wrapper(*args, **kwargs):
        if not started():
            raise RuntimeError(f'You need to call pt.java.init() required before you can call {fn}')
        return fn(*args, **kwargs)
    return typing.cast(T, _wrapper) # noqa: PT100 (this is typing.cast, not jinus.cast)


def before_init(fn: T) -> T:
    """
    If the JVM has already started, an error is raised.
    """
    @wraps(fn)
    def _wrapper(*args, **kwargs):
        if started():
            raise RuntimeError(f'You can only call {fn} before either you start using java or call pt.java.init()')
        return fn(*args, **kwargs)
    return typing.cast(T, _wrapper) # noqa: PT100 (this is typing.cast, not jinus.cast)


# ----------------------------------------------------------
# Jnius Wrappers
# ----------------------------------------------------------
# These functions wrap jnius to make sure that java is
# running before they're called. Doing it this way allows
# functions to import them before java is loaded.
# ----------------------------------------------------------

@required_raise
def autoclass(*args, **kwargs):
    """
    Wraps jnius.autoclass once java has started. Raises an error if called before pt.java.init() is called.
    """
    import jnius # noqa: PT100
    return jnius.autoclass(*args, **kwargs) # noqa: PT100


@required_raise
def cast(*args, **kwargs):
    """
    Wraps jnius.cast once java has started. Raises an error if called before pt.java.init() is called.
    """
    import jnius # noqa: PT100
    return jnius.cast(*args, **kwargs) # noqa: PT100


# ----------------------------------------------------------
# Init
# ----------------------------------------------------------
# This function (along with legacy_init) loads all modules
# registered via pyterrier.java.init entry points and starts
# the JVM.
# ----------------------------------------------------------

def init() -> None:
    _init()


@pt.utils.once()
def _init(trigger=None):
    global _started
    # java/jnius is now optional, should check here that it's installed. E.g.,
    from importlib.util import find_spec
    if find_spec('jnius_config') is None:
        raise ImportError(f'pyterrier[java] not installed; cannot run pt.java.init(), triggered by {trigger}. Run `pip install pyterrier[java]` to install it.')
        
    # TODO: what about errors during init? What happens to _started? Etc.
    initializers = []
    for entry_point in pt.utils.entry_points('pyterrier.java.init'):
        initalizer = entry_point.load()()
        if initalizer.condition():
            initializers.append((entry_point.name, initalizer))

    if len(initializers) == 0:
        raise RuntimeError('No Java initializers found. This is likely a configuration issue with the package. '
                           'If installed using `pip install -e .` or `python setup.py develop`, try reinstalling.')

    # sort by priority
    initializers = sorted(initializers, key=lambda i: i[1].priority())

    import jnius_config

    # run pre-initialization setup
    for _, initializer in initializers:
        initializer.pre_init(jnius_config)

    import jnius # noqa: PT100 
    _started = True

    # run post-initialization setup
    for _, initializer in initializers:
        initializer.post_init(jnius)

    # build "Java started" message
    message = []
    if trigger:
        message.append(f'Java started (triggered by {trigger}) and loaded: ')
    else:
        message.append('Java started and loaded: ')
    for name, initializer in initializers:
        msg = initializer.message()
        if msg is None:
            message.append(name)
        else:
            message.append(f'{name} [{msg}]')
    sys.stderr.write(message[0] + ', '.join(message[1:]) + '\n')


@before_init
def legacy_init(version=None, mem=None, packages=[], jvm_opts=[], redirect_io=True, logging='WARN', home_dir=None, boot_packages=[], tqdm=None, no_download=False, helper_version = None):
    """
    Function that can be called before Terrier classes and methods can be used.
    Loads the Terrier .jar file and imports classes. Also finds the correct version of Terrier to download if no version is specified.

    Args:
        version(str): Which version of Terrier to download. Default is `None`.

         * If None, find the newest Terrier released version in MavenCentral and download it.
         * If `"snapshot"`, will download the latest build from `Jitpack <https://jitpack.io/>`_.

        mem(str): Maximum memory allocated for the Java virtual machine heap in MB. Corresponds to java `-Xmx` commandline argument. Default is 1/4 of physical memory.
        boot_packages(list(str)): Extra maven package coordinates files to load before starting Java. Default=`[]`. There is more information about loading packages in the `Terrier documentation <https://github.com/terrier-org/terrier-core/blob/5.x/doc/terrier_develop.md>`_
        packages(list(str)): Extra maven package coordinates files to load, using the Terrier classloader. Default=`[]`. See also `boot_packages` above.
        jvm_opts(list(str)): Extra options to pass to the JVM. Default=`[]`. For instance, you may enable Java assertions by setting `jvm_opts=['-ea']`
        redirect_io(boolean): If True, the Java `System.out` and `System.err` will be redirected to Pythons sys.out and sys.err. Default=True.
        logging(str): the logging level to use:

         * Can be one of `'INFO'`, `'DEBUG'`, `'TRACE'`, `'WARN'`, `'ERROR'`. The latter is the quietest.
         * Default is `'WARN'`.

        home_dir(str): the home directory to use. Default to PYTERRIER_HOME environment variable.
        tqdm: The `tqdm <https://tqdm.github.io/>`_ instance to use for progress bars within PyTerrier. Defaults to tqdm.tqdm. Available options are `'tqdm'`, `'auto'` or `'notebook'`.
        helper_version(str): Which version of the helper.

   
    **Locating the Terrier .jar file:** PyTerrier is not tied to a specific version of Terrier and will automatically locate and download a recent Terrier .jar file. However, inevitably, some functionalities will require more recent Terrier versions. 
    
     * If set, PyTerrier uses the `version` init kwarg to determine the .jar file to look for.
     * If the `version` init kwarg is not set, Terrier will query MavenCentral to determine the latest Terrier release.
     * If `version` is set to `"snapshot"`, the latest .jar file build derived from the `Terrier Github repository <https://github.com/terrier-org/terrier-core/>`_ will be downloaded from `Jitpack <https://jitpack.io/>`_.
     * Otherwise the local (`~/.mvn`) and MavenCentral repositories are searched for the jar file at the given version.
    In this way, the default setting is to download the latest release of Terrier from MavenCentral. The user is also able to use a locally installed copy in their private Maven repository, or track the latest build of Terrier from Jitpack.
    
    If you wish to run PyTerrier in an offline enviroment, you should ensure that the "terrier-assemblies-{your version}-jar-with-dependencies.jar" and "terrier-python-helper-{your helper version}.jar"
    are in the  "~/.pyterrier" (if they are not present, they will be downloaded the first time). Then you should set their versions when calling ``init()`` function. For example:
    ``pt.init(version = 5.5, helper_version = "0.0.6")``.
    """

    deprecated_calls = []

    # Set the corresponding options
    if mem is not None:
        pt.java.set_memory_limit(mem)
        deprecated_calls.append(f'pt.java.set_memory_limit({mem!r})')

    if not redirect_io:
        pt.java.set_redirect_io(redirect_io)
        deprecated_calls.append(f'pt.java.set_redirect_io({redirect_io!r})')

    if logging != 'WARN':
        pt.java.set_log_level(logging)
        deprecated_calls.append(f'pt.java.set_log_level({logging!r})')

    for package in boot_packages:
        # format: org:package:version:filetype (where version and filetype are optional)
        pkg_split = package.split(':')
        pkg_string = ", ".join(f'{w!r}' for w in pkg_split)
        pt.java.add_package(*pkg_split) 
        deprecated_calls.append(f'pt.java.add_package({pkg_string})')

    for opt in jvm_opts:
        pt.java.add_option(opt)
        deprecated_calls.append(f'pt.java.add_option({opt!r})')

    if version is not None:
        pt.terrier.set_version(version)
        deprecated_calls.append(f'pt.terrier.set_version({version!r})')

    if helper_version is not None:
        pt.terrier.set_helper_version(helper_version)
        deprecated_calls.append(f'pt.terrier.set_helper_version({helper_version!r})')

    if tqdm is not None:
        pt.utils.set_tqdm(tqdm)
        deprecated_calls.append(f'pt.utils.set_tqdm({tqdm!r})')

    if no_download:
        pt.java.mavenresolver.offline()
        deprecated_calls.append('pt.java.mavenresolver.offline()')

    pt.java.init()
    deprecated_calls.append('pt.java.init() # optional, forces java initialisation')

    # Import other java packages
    if packages:
        pkgs_string = ",".join(packages)
        pt.terrier.set_property("terrier.mvn.coords", pkgs_string)
        deprecated_calls.append(f'pt.terrier.set_property("terrier.mvn.coords", {pkgs_string!r})')

    # Warning to give new initialization
    deprecated_message = 'Call to deprecated method pt.init(). Deprecated since version 0.11.0.'
    if len(deprecated_calls) > 1: # called setup other than pt.java.init()
        deprecated_message = deprecated_message + '\nThe following code will have the same effect:'
    else: # only called pt.java.init()
        deprecated_message = deprecated_message + '\njava is now started automatically with default settings. To force initialisation early, run:'
    deprecated_message = '\n'.join([deprecated_message] + deprecated_calls)
    with warnings.catch_warnings():
        warnings.simplefilter('once', DeprecationWarning) # DeprecationWarning hidden by default, @deprecated does this to show the messages
        warnings.warn(deprecated_message, category=DeprecationWarning, stacklevel=3) # stacklevel=3 prints wherever this call comes from, rather than here, see @deprecated source


def started() -> bool:
    """
    Returns True if pt.java.init() has been called. Otherwise False.
    """
    return _started


class JavaInitializer:
    """
    A `JavaInitializer` manages the initilization of a module that uses java components. The two main methods are
    `pre_init` and `post_init`, which perform configuration before and after the JVM has started, respectively.
    """

    def priority(self) -> int:
        """
        Returns the priority of this initializer. A lower priority is executed first.
        """
        return 0

    def condition(self) -> bool:
        """
        Returns True if the initializer should be run. Otherwise False.
        """
        return True

    def pre_init(self, jnius_config) -> None:
        """
        Called before the JVM is started. `jnius_config` is the `jnius_config` module, whic can be used to configure
        java, such as by adding jars to the classpath.
        """
        pass

    def post_init(self, jnius) -> None:
        """
        Called after the JVM has started. `jnius` is the `jnius` module, which can be used to interact with java.
        """
        pass

    def message(self) -> Optional[str]:
        """
        Returns a message to be displayed after the JVM has started alongside the name of the entry point. If None,
        only the entry point name will be displayed.
        """
        return None


# ----------------------------------------------------------
# Parallel
# ----------------------------------------------------------
# These functions are for working in parallel mode, e.g.,
# with multiprocessing. They help restarting and configure
# the JVM the same way it was when it was started in the
# parent process
# ----------------------------------------------------------

def parallel_init(started: bool, configs: Dict[str, Dict[str, Any]]) -> None:
    global _configs
    if started:
        if not pt.java.started():
            _configs = configs
            _init(trigger='parallel_init')
        else:
            warnings.warn(
                "Avoiding reinit of PyTerrier")


def parallel_init_args() -> Tuple[bool, Dict[str, Dict[str, Any]]]:
    return (
        started(),
        deepcopy(_configs),
    )


# ----------------------------------------------------------
# Configuration Utils
# ----------------------------------------------------------
# We need a global store of all java-related configurations
# so that when running in parallel, we can set everying back
# up the same way it started. These utils help manage this
# global configuration.
# ----------------------------------------------------------

class Configuration:
    def __init__(self, name):
        self.name = name

    def get(self, key):
        return deepcopy(_configs[self.name][key])

    @before_init
    def set(self, key, value):
        self(**{key: value})

    def append(self, key, value):
        res = self.get(key)
        res.append(value)
        self(**{key: res})

    def __getitem__(self, key):
        return self.get(key)

    @before_init
    def __setitem__(self, key, value):
        self.set(key, value)

    def __call__(self, **settings: Any):
        if started() and any(settings):
            raise RuntimeError('You cannot change java settings after java has started')
        for key, value in settings.items():
            if key not in _configs[self.name]:
                raise AttributeError(f'{key!r} not defined as a java setting for {self.name!r}')
            _configs[self.name][key] = value
        return deepcopy(_configs[self.name])


def register_config(name, config: Dict[str, Any]):
    assert name not in _configs
    _configs[name] = deepcopy(config)
    return Configuration(name)


# ----------------------------------------------------------
# Java Classes
# ----------------------------------------------------------
# This class enables the lazy loading of java classes. It
# helps avoid needing a ton of autoclass() statements to
# pre-load Java classes.
# ----------------------------------------------------------

class JavaClasses:
    def __init__(self, **mapping: Union[str, Callable[[], str]]):
        self._mapping = mapping
        self._cache : Dict[str, Callable]= {}

    def __dir__(self):
        return list(self._mapping.keys())

    @required_raise
    def __getattr__(self, key: str) -> Any:
        if key not in self._mapping:
            return AttributeError(f'{self} has no attribute {key!r}')
        if key not in self._cache:
            clz = self._mapping[key]
            if callable(clz):
                clz = clz()
            self._cache[key] = pt.java.autoclass(clz)
        return self._cache[key]



import pyterrier as pt

import itertools
import unittest
import tempfile
import shutil
import os

from .base import TempDirTestCase, BaseTestCase

class TestIterDictIndexer(TempDirTestCase):
        

    def _create_index(self, it, indexer):
        print("Writing index to " + self.test_dir)
        indexref = indexer.index(it)
        self.assertIsNotNone(indexref)
        return indexref

    def _make_check_index(self, n, index_type, attrs=('text',), meta=('docno', 'url', 'title')):
        from pyterrier.terrier.index import IndexingType
        # Test both versions: _fifo (for UNIX) and _nofifo (for Windows)
        indexers = [
            pt.index._IterDictIndexer_fifo(self.test_dir, text_attrs=attrs, type=index_type, meta=meta, fields=True),
            pt.index._IterDictIndexer_fifo(self.test_dir, text_attrs=attrs, type=index_type, threads=4, meta=meta, fields=True),
            pt.index._IterDictIndexer_nofifo(self.test_dir, text_attrs=attrs, type=index_type, meta=meta, fields=True),
        ]
        if pt.utils.is_windows():
           indexers = [indexers[-1]] 
        for indexer in indexers:
            with self.subTest(indexer=indexer):
                it = (
                    {'docno': '1', 'url': 'url1', 'text': 'He ran out of money, so he had to stop playing', 'title': 'Woes of playing poker', 'raw_source': b'<some>xml</content>'},
                    {'docno': '2', 'url': 'url2', 'text': 'The waves were crashing on the shore; it was a', 'title': 'Lovely sight', 'raw_source': b'<some>xml</content>'},
                    {'docno': '3', 'url': 'url3', 'text': 'The body may perhaps compensates for the loss', 'title': 'Best of Viktor Prowoll', 'raw_source': b'<some>xml</content>'},
                )
                it = itertools.islice(it, n)
                indexref = self._create_index(it, indexer)
                index = pt.IndexFactory.of(indexref)
                self.assertIsNotNone(index)
                self.assertEqual(n, index.getCollectionStatistics().getNumberOfDocuments())
                self.assertEqual(len(set(meta)), len(index.getMetaIndex().getKeys()))
                for m in meta:
                    self.assertTrue(m in index.getMetaIndex().getKeys())
                self.assertEqual(len(attrs), index.getCollectionStatistics().numberOfFields)
                for f in attrs:
                    self.assertTrue(f in index.getCollectionStatistics().getFieldNames())
                if index_type is IndexingType.CLASSIC:
                    self.assertTrue(os.path.isfile(self.test_dir + '/data.direct.bf'))
                    self.assertTrue(index.hasIndexStructure("direct"))
                else:
                    self.assertFalse(os.path.isfile(self.test_dir + '/data.direct.bf'))
                inv = index.getInvertedIndex()
                lex = index.getLexicon()
                post = inv.getPostings(lex.getLexiconEntry('plai'))
                post.next()
                from jnius import cast
                post = cast("org.terrier.structures.postings.FieldPosting", post)
                if 'title' in attrs:
                    self.assertEqual(2, post.frequency)
                    self.assertEqual(1, post.fieldFrequencies[0])
                    self.assertEqual(1, post.fieldFrequencies[1])
                else:
                    self.assertEqual(1, post.frequency)
                    self.assertEqual(1, post.fieldFrequencies[0])
                index.close()
            # reset index directory for next run
            shutil.rmtree(self.test_dir)
            os.mkdir(self.test_dir)

    def test_createindex1_2fields(self):
        from pyterrier.terrier.index import IndexingType
        self._make_check_index(1, IndexingType.CLASSIC, attrs=['text', 'title'])

    def test_createindex2_2fields(self):
        from pyterrier.terrier.index import IndexingType
        self._make_check_index(2, IndexingType.CLASSIC, attrs=['text', 'title'])

    def test_createindex3_2fields(self):
        from pyterrier.terrier.index import IndexingType
        self._make_check_index(3, IndexingType.CLASSIC, attrs=['text', 'title'])

    def test_createindex1_single_pass_2fields(self):
        from pyterrier.terrier.index import IndexingType
        self._make_check_index(1, IndexingType.SINGLEPASS, attrs=['text', 'title'])

    def test_createindex2_single_pass_2fields(self):
        from pyterrier.terrier.index import IndexingType
        self._make_check_index(2, IndexingType.SINGLEPASS, attrs=['text', 'title'])

    def test_createindex3_single_pass_2fields(self):
        from pyterrier.terrier.index import IndexingType
        self._make_check_index(3, IndexingType.SINGLEPASS, attrs=['text', 'title'])


if __name__ == "__main__":
    unittest.main()



import os
import json
import pandas as pd
from abc import ABC, abstractmethod
from typing import Tuple, Iterable, List, Optional, Union, Protocol, runtime_checkable, Dict, Any
from collections import defaultdict
import zipfile
import tarfile
import types
import requests
import urllib
import functools
from warnings import warn
import pyterrier as pt
from pyterrier.transformer import is_lambda


# DATASET_MAP provides legacy support for registering datasets in extensions. It will be removed in a future version
# in favor of using a DatasetProvider registered with an entry point.
DATASET_MAP: Dict[str, Any] = {}


class Dataset:
    """Represents a dataset (test collection) for indexing and/or retrieval.

    A common use case is for an Experiment::

        dataset = pt.get_dataset("trec-robust-2004")
        pt.Experiment([br1, br2], dataset.get_topics(), dataset.get_qrels(), eval_metrics=["map", "recip_rank"])
    """
    def get_corpus(self) -> List[str]:
        """Returns the location of the files to allow indexing the corpus, i.e. it returns a list of filenames.

        .. deprecated::
            Use :meth:`get_corpus_iter` instead.

        Raises:
            NotImplementedError: If the dataset does not support this method.
        """
        raise NotImplementedError(f'{self!r} does not support get_corpus')

    def get_corpus_iter(self, *, verbose: bool = True) -> pt.model.IterDict:
        """Returns an iter of dicts for this collection.

        Args:
            verbose (bool, optional): Whether to print progress bar. Defaults to True.

        Raises:
            NotImplementedError: If the dataset does not support this method.
        """
        raise NotImplementedError(f'{self!r} does not support get_corpus_iter')

    def get_corpus_lang(self) -> Optional[str]:
        """Returns the ISO 639-1 language code for the corpus, or None for multiple/other/unknown."""
        return None

    def get_index(self, variant=None, **kwargs): # TODO: -> pt.Artifact
        """Returns the IndexRef of the index to allow retrieval. Only a few datasets provide indices ready made.

        Raises:
            NotImplementedError: If the dataset does not support this method.
        """
        raise NotImplementedError(f'{self!r} does not support get_index')

    def get_topics(self, variant: Optional[str] = None) -> pd.DataFrame:
        """Returns the topics (if available), as a dataframe, ready for retrieval. 

        Args:
            variant (str, optional): The variant of the topics, such as a sub-dataset of the field. Defaults to None.

        Raises:
            NotImplementedError: If the dataset does not support this method.
        """
        raise NotImplementedError(f'{self!r} does not support get_topics')

    def get_topics_lang(self) -> Optional[str]:
        """Returns the ISO 639-1 language code for the topics, or None for multiple/other/unknown."""
        return None

    def get_qrels(self, variant: Optional[str] = None) -> pd.DataFrame:
        """Returns the qrels, as a dataframe, for evaluation (if available).

        Args:
            variant (str, optional): The variant of the topics, such as a sub-dataset of the field. Defaults to None.

        Raises:
            NotImplementedError: If the dataset does not support this method.
        """
        raise NotImplementedError(f'{self!r} does not support get_qrels')

    def get_topicsqrels(self, variant: Optional[str] = None) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Returns both the topics and qrels in a tuple. This is useful for pt.Experiment().

        Args:
            variant (str, optional): The variant of the topics, such as a sub-dataset of the field. Defaults to None.

        Raises:
            NotImplementedError: If the dataset does not support this method.
        """
        return (
            self.get_topics(variant=variant),
            self.get_qrels(variant=variant)
        )

    def info_url(self) -> Optional[str]:
        """Returns a url that provides more information about this dataset, or None if URL is unknown."""
        return None

    def get_results(self, variant: Optional[str] = None) -> pd.DataFrame:
        """Returns a standard result set provided by the dataset (if available). This is useful for re-ranking experiments.

        Args:
            variant (str, optional): The variant of the topics, such as a sub-dataset of the field. Defaults to None.

        Raises:
            NotImplementedError: If the dataset does not support this method.
        """
        raise NotImplementedError(f'{self!r} does not support get_results')

    def _describe_component(self, component: str) -> Optional[Union[bool, List[str]]]:
        return None


class DatasetProvider(ABC):
    """ Represents a source of datasets. For instance: the default suite of datasets provided directly by PyTerrier,
    or datasets from the ``ir-datasets`` package.

    Dataset providers should be registred as ``pyterrier.dataset_provider`` entry points.
    """

    @abstractmethod
    def get_dataset(self, name: str) -> Dataset:
        """ Returns a dataset for the provided identifier (``name``).

        Args:
            name (str): The identifier of the dataset.

        Returns:
            Dataset: The dataset.

        Raises:
            KeyError: If the dataset is not found.
        """

    def list_dataset_names(self) -> Iterable[str]:
        """ Returns the names of the datasets that this provider gives access to.

        The primary purpose of this method is to populate the `list of datasets <https://pyterrier.readthedocs.io/en/latest/datasets.html>`__
        in the documentation.

        .. note::

            This method is optional and does not need to return all available datasets for this provider. For instance,
            sometimes the number of datasets is too large to reasonably display, or it may be too expensive to fetch
            all datasets.

        Returns:
            Iterable[str]: An iterable of tuples, where the first element is the name of the dataset and the second
                element is the dataset itself.
        """
        return []


class RemoteDataset(Dataset):

    def __init__(self, name, locations):
        self.locations = locations
        self.name = name
        self.user = None
        self.password = None

    def _configure(self, user: Optional[str] = None, password: Optional[str] = None):
        self.corpus_home = os.path.join(pt.io.pyterrier_home(), "corpora", self.name)
        if user is not None:
            self.user = user
        if password is not None:
            self.password = password

    @staticmethod
    def download(URLs : Union[str,List[str]], filename : str, **kwargs):
        basename = os.path.basename(filename)

        if isinstance(URLs, str):
            URLs = [URLs]
        
        finalattempt=len(URLs)-1
        error = None
        for i, url in enumerate(URLs):            
            try:
                r = requests.get(url, allow_redirects=True, stream=True, **kwargs)
                r.raise_for_status()
                total = int(r.headers.get('content-length', 0))
                with pt.io.finalized_open(filename, 'b') as file, pt.tqdm( # type: ignore
                        desc=basename,
                        total=total,
                        unit='iB',
                        unit_scale=True,
                        unit_divisor=1024,
                ) as bar:
                    for data in r.iter_content(chunk_size=1024):
                        size = file.write(data)
                        bar.update(size)
                    break
            except Exception as e:
                if error is not None:
                    e.__cause__ = error # chain errors to show all if fails
                error = e
                if i == finalattempt:
                    raise error
                else:
                    warn(
                        "Problem fetching %s, resorting to next mirror" % url)
            

    def _check_variant(self, component, variant=None):
        name = self.name
        if component not in self.locations:
            raise ValueError("No %s in dataset %s" % (component, name))
        if variant is None:
            if not isinstance(self.locations[component], list):
                raise ValueError("For %s in dataset %s, you must specify a variant. Available are: %s" % (component, name, str(list(self.locations[component].keys()))))
        else:
            if isinstance(self.locations[component], list):
                raise ValueError("For %s in dataset %s, there are no variants, but you specified %s" % (component, name, variant))
            if variant not in self.locations[component]:
                raise ValueError("For %s in dataset %s, there is no variant %s. Available are: %s" % (component, name, variant, str(list(self.locations[component].keys()))))

    def _get_one_file(self, component, variant=None):
        filetype=None
        name=self.name
        self._check_variant(component, variant)
        location = self.locations[component][0] if variant is None else self.locations[component][variant]

        if is_lambda(location) or isinstance(location, types.FunctionType):
            argcount = location.__code__.co_argcount
            if argcount == 0:
                return location()
            elif argcount == 3:
                return location(self, component, variant)
            else:
                raise TypeError("Expected function with 0 or 3 arguments for  %s %s %s" % (component, name, variant))

        local = location[0]
        URL = location[1]
        if len(location) > 2:
            filetype = location[2]

        if not os.path.exists(self.corpus_home):
            os.makedirs(self.corpus_home)

        local = os.path.join(self.corpus_home, local)
        actualURL = URL if isinstance(URL, str) else URL[0]
        if "#" in actualURL and not os.path.exists(local):
            tarname, intarfile = actualURL.split("#")
            assert "/" not in intarfile
            assert ".tar" in tarname or ".tgz" in tarname or ".zip" in tarname
            localtarfile, _ = self._get_one_file("tars", tarname)
            extractor = zipfile.ZipFile if ".zip" in tarname else tarfile.open
            with extractor(localtarfile, "r") as tarobj:
                tarobj.extract(intarfile, path=self.corpus_home)
            os.rename(os.path.join(self.corpus_home, intarfile), local)
            return (local, filetype)

        if not os.path.exists(local):
            try:
                print("Downloading %s %s to %s" % (self.name, component, local))
                kwargs = {}
                if self.user is not None:
                    kwargs["auth"]=(self.user, self.password)
                RemoteDataset.download(URL, local, **kwargs)
            except urllib.error.HTTPError as he:
                raise ValueError("Could not fetch " + URL) from he
        return (local, filetype)

    def _get_all_files(self, component, variant=None, **kwargs):
        if variant is None:
            localDir = os.path.join(self.corpus_home, component)
        else:
            localDir = os.path.join(self.corpus_home, component, variant)

        kwargs = {}
        if self.user is not None:
            kwargs["auth"]=(self.user, self.password)

        direxists = os.path.exists(localDir)
        
        location = self.locations[component]
        if is_lambda(location) or isinstance(location, types.FunctionType):
            # functions are expensive to call, normally another HTTP is needed.
            # just assume we have everthing we need if we have the local directory already
            # and it contains a .complete file.
            if direxists and os.path.exists(os.path.join(localDir, ".complete")):
                return localDir

            # call the function, and get the file list
            file_list = location(self, component, variant, **kwargs)
        else:
            file_list = self.locations[component] if variant is None else self.locations[component][variant]

        if not direxists:
            os.makedirs(localDir)
            print("Downloading %s %s to %s" % (self.name, component, localDir))
        

        # check for how much space is required and available space
        def _totalsize(file_list):
            total = -1
            for f in file_list:
                if len(f) > 2:
                    total += f[2]
            if total != -1:
                total += 1
            return total

        totalsize = _totalsize(file_list)
        if totalsize > 0:
            import shutil
            total, used, free = shutil.disk_usage(localDir)
            if free < totalsize:
                raise ValueError("Insufficient freedisk space at %s to download index" % localDir)
            if totalsize > 2 * 2**30:
                warn(
                    "Downloading index of > 2GB.")

        # all tarfiles that we will need to process        
        tarfiles = defaultdict(list)
        for fileentry in file_list:
            local = fileentry[0]
            URL = fileentry[1]
            assert "/" not in local, "cant handle / in %s, local name is %s" % (local)
            expectedlength = -1
            if len(fileentry) == 3:
                expectedlength = fileentry[2]
            local = os.path.join(localDir, local)
            
            # if file exists and we know length, check if dowload is complete
            fileexists = os.path.exists(local)
            if fileexists and expectedlength >= 0:
                length = os.stat(local).st_size
                if expectedlength != length:
                    warn(
                        "Removing partial download of %s (expected %d bytes, found %d)" % (local, expectedlength, length ))
                    os.remove(local)
                    fileexists = False

            if not fileexists:
                if "#" in URL:
                    tarname, intarfile = URL.split("#")
                    assert ".tar" in tarname or ".tgz" in tarname or ".zip" in tarname, "I dont know how to decompress file %s" % tarname
                    localtarfile, _ = self._get_one_file("tars", tarname)
                    # append intarfile to the list of files to be extracted from localtarfile
                    tarfiles[localtarfile].append((intarfile, local))
                else:
                    try:
                        RemoteDataset.download(URL, local, **kwargs)
                    except urllib.error.HTTPError as he:
                        raise ValueError("Could not fetch " + URL) from he

                    # verify file if exists
                    if expectedlength >= 0:
                        length = os.stat(local).st_size
                        if expectedlength != length:
                            raise ValueError("Failed download of %s to %s (expected %d bytes, found %d)" % (URL, local, expectedlength, length ))

        # now extract all required files from each tar file
        for localtarfile in tarfiles:
            extractor = zipfile.ZipFile if ".zip" in tarname else tarfile.open
            with extractor(localtarfile, "r") as tarobj:
                # 5 is abrtary threshold - if we have lots of files to extract, give a progress bar. alternative would be delay=5?
                iter = pt.tqdm(tarfiles[localtarfile], unit="file", desc="Extracting from " + localtarfile) if len(tarfiles[localtarfile]) > 5 else tarfiles[localtarfile]
                for (intarfile, local) in iter:
                    tarobj.extract(intarfile, path=self.corpus_home)
                    local = os.path.join(self.corpus_home, local)
                    os.rename(os.path.join(self.corpus_home, intarfile), local)
                    #TODO, files /could/ be recompressed here to save space, if not already compressed

        # finally, touch a file signifying that download has been completed
        pt.io.touch(os.path.join(localDir, ".complete"))
        return localDir

    def _describe_component(self, component):
        if component not in self.locations:
            return None
        if isinstance(self.locations[component], list):
            return True
        if isinstance(self.locations[component], dict):
            return list(self.locations[component].keys())
        return True

    def get_corpus(self, **kwargs):
        return list(filter(lambda f : not f.endswith(".complete"), pt.io.find_files(self._get_all_files("corpus", **kwargs))))

    def get_corpus_iter(self, **kwargs):
        if "corpus_iter" not in self.locations:
            raise NotImplementedError(f"{self!r} does not support get_corpus_iter")
        return self.locations["corpus_iter"](self, **kwargs)
        
    def get_corpus_lang(self):
        if 'corpus' in self.locations:
            return 'en' # all are english
        return None

    def get_qrels(self, variant=None):
        filename, type = self._get_one_file("qrels", variant)
        if type == "direct":
            return filename 
        return pt.io.read_qrels(filename)

    def get_topics(self, variant=None, **kwargs):
        file, filetype = self._get_one_file("topics", variant)
        if filetype is None or filetype in pt.io.SUPPORTED_TOPICS_FORMATS:
            return pt.io.read_topics(file, format=filetype, **kwargs)
        elif filetype == "direct":
            return file
        raise ValueError("Unknown filetype %s for %s topics %s"  % (filetype, self.name, variant))
    
    def get_topics_lang(self):
        if 'topics' in self.locations:
            return 'en' # all are english
        return None

    def get_index(self, variant=None, **kwargs):
        thedir = self._get_all_files("index", variant=variant, **kwargs)
        return thedir

    def __repr__(self):
        return f"RemoteDataset({self.name!r}, { {k: ... for k in self.locations}!r})"

    def info_url(self):
        return self.locations.get('info_url')


@runtime_checkable
class HasConfigure(Protocol):
    def _configure(self, **kwargs):
        pass

def add_tokenize_query_arg(fn):
    @functools.wraps(fn)
    def _wrapper(variant: Optional[str] = None, tokenise_query: Optional[bool] = None):
        topics = fn(variant)
        if topics is not None and tokenise_query is not None and 'query' in topics:
            raise ValueError("get_topics(tokenise_query:bool) has been removed. All datasets queries are now untokenised by default, and terrier.Retriever will handle them as needed.")
        return topics
    return _wrapper
_loaded_providers = {}


def _load_all_providers():
    for ep in pt.utils.entry_points('pyterrier.dataset_provider'):
        _loaded_providers[ep.name] = ep.load()()


def _load_provider(name: str) -> Optional[DatasetProvider]:
    if name not in _loaded_providers:
        for ep in pt.utils.entry_points('pyterrier.dataset_provider'):
            if ep.name == name:
                _loaded_providers[name] = ep.load()()
                break
    return _loaded_providers.get(name)


def get_dataset(name: str, **configure_kwargs) -> Dataset:
    """Get a dataset by name.

    The name may provide a provider as a prefix in the format of ``provider:dataset``. In this case, the dataset is
    loaded from that provider (e.g., ``irds``).

    Args:
        name (str): The name of the dataset.
        **configure_kwargs: Additional configuration arguments to pass to the dataset's ``_configure()`` method (if one is available)

    Returns:
        Dataset: The dataset.

    Raises:
        KeyError: If the dataset is not found.
    """
    if name in DATASET_MAP:
        result = DATASET_MAP[name]
    else:
        orig_name = name
        provider_name = 'builtin' # default provider when no provider: prefix is given
        if ':' in name:
            provider_name, name = name.split(':', 1)

        provider = _load_provider(provider_name)

        if provider is None:
            if provider_name == 'builtin':
                # builtin should always be found; an absence of builtin means the pyterrier is not installed properly
                raise KeyError(f'Dataset {orig_name!r} not found due to missing provider {provider_name!r}. '
                               +'You need to pip install python-terrier to ensure dataset providers can be found.')
            # We could provide suggestions here on packages to install for missing entry points in due course
            raise KeyError(f'Dataset {orig_name!r} not found due to missing provider {provider_name!r}. Are you missing a package?')

        try:
            result = provider.get_dataset(name)
        except KeyError as ex:
            raise KeyError(f'Dataset {orig_name!r} not found in provider {provider_name!r}.') from ex

    # Configure
    if isinstance(result, HasConfigure):
        result._configure(**configure_kwargs)
    elif configure_kwargs:
        raise TypeError(f'Unsupported keyword arguments passed to get_dataset: {get_dataset}')
    
    # Temporary handling of topic tokenization ... we should remove this at some point, but let's keep it now so it provides a clearer reason
    # to uswers why it now breaks.
    if not hasattr(result.get_topics, '_has_add_tokenize_query_arg_applied'):
        result.get_topics = add_tokenize_query_arg(result.get_topics)    
    return result


def datasets() -> List[str]:
    """Lists the names of available datasets."""
    _load_all_providers()
    result = list(DATASET_MAP.keys())
    for provider_name in sorted(_loaded_providers.keys()):
        provider = _loaded_providers[provider_name]
        for name in provider.list_dataset_names():
            if provider_name != '' and provider_name != 'builtin':
                name = f'{provider_name}:{name}'
            result.append(name)
    return result


def find_datasets(query, en_only=True):
    """
    A grep-like method to help identify datasets. Filters the output of list_datasets() based on the name containing the query
    """
    datasets = list_datasets(en_only=en_only)
    return datasets[datasets['dataset'].str.contains(query)]


def list_datasets(en_only=True):
    """
        Returns a dataframe of all datasets, listing which topics, qrels, corpus files or indices are available.
        By default, filters to only datasets with both a corpus and topics in English.
    """
    with pt.utils.temp_env("IR_DATASETS_SKIP_DEPRECATED_WARNING", "true"): # hide IRDS deprecated warning when loading all datasets
        rows = []
        for k in datasets():
            dataset = get_dataset(k)
            rows.append([
                k, 
                dataset._describe_component("topics"), 
                dataset.get_topics_lang(), 
                dataset._describe_component("qrels"), 
                dataset._describe_component("corpus"), 
                dataset.get_corpus_lang(), 
                dataset._describe_component("index"), 
                dataset.info_url()
            ])
    
    result = pd.DataFrame(rows, columns=["dataset", "topics", "topics_lang", "qrels", "corpus", "corpus_lang", "index", "info_url"])
    if en_only:
        topics_filter = (result['topics'].isnull()) | (result['topics_lang'] == 'en')
        corpus_filter = (result['corpus'].isnull()) | (result['corpus_lang'] == 'en')
        result = result[topics_filter & corpus_filter]
    return result


def transformer_from_dataset(
    dataset : Union[str, Dataset],
    clz,
    variant: Optional[str] = None,
    version: str = 'latest',        
    **kwargs
) -> pt.Transformer:
    """Returns a Transformer instance of type ``clz`` for the provided index of variant ``variant``."""
    if isinstance(dataset, str):
        dataset = get_dataset(dataset)
    if version != "latest":
        raise ValueError("index versioning not yet supported")
    indexref = dataset.get_index(variant)

    classname = clz.__name__
    classnames = [classname]
    if classname == 'Retriever':
        # we need to look for BatchRetrieve.args.json for legacy support
        classnames.append('BatchRetrieve')
    for c in classnames:
        # now look for, e.g., BatchRetrieve.args.json file, which will define the args for Retriever, e.g. stemming
        indexdir = indexref #os.path.dirname(indexref.toString())
        argsfile = os.path.join(indexdir, classname + ".args.json")
        if os.path.exists(argsfile):
            with pt.io.autoopen(argsfile, "rt") as f:
                args = json.load(f)
                # anything specified in kwargs of this methods overrides the .args.json file
                args.update(kwargs)
                kwargs = args
    return clz(indexref, **kwargs)



import numpy as np
import pandas as pd
from collections import namedtuple
from . import MEASURES_TYPE
from ir_measures import Measure
from typing import Sequence, Tuple, Dict, Union, List, Optional, overload, Literal, Any

class EvaluationDataTuple(namedtuple('EvaluationDataTuple', ['averages', 'perquery'])):
    averages : pd.DataFrame
    perquery : pd.DataFrame

    def _repr_html_(self, *a, **kw):
        return self.averages._repr_html_(*a, *kw) + self.perquery._repr_html_(*a, *kw)

def _bold_cols(data : pd.Series, col_type):
    if data.name not in col_type:
        return [''] * len(data)

    colormax_attr = 'font-weight: bold'
    colormaxlast_attr = 'font-weight: bold'
    if col_type[data.name] == "+":
        max_value = data.max()
    else:
        max_value = data.min()

    is_max = [colormax_attr if v == max_value else '' for v in data]
    is_max[len(data) - list(reversed(data)).index(max_value) -  1] = colormaxlast_attr
    return is_max


def _color_cols(data : pd.Series, col_type,
                       colormax='antiquewhite', colormaxlast='lightgreen',
                       colormin='antiquewhite', colorminlast='lightgreen' ):
    if data.name not in col_type:
      return [''] * len(data)

    if col_type[data.name] == "+":
      colormax_attr = f'background-color: {colormax}'
      colormaxlast_attr = f'background-color: {colormaxlast}'
      max_value = data.max()
    else:
      colormax_attr = f'background-color: {colormin}'
      colormaxlast_attr = f'background-color: {colorminlast}'
      max_value = data.min()

    is_max = [colormax_attr if v == max_value else '' for v in data]
    is_max[len(data) - list(reversed(data)).index(max_value) -  1] = colormaxlast_attr
    return is_max


def _mean_of_measures(result, measures=None, num_q = None):
    if len(result) == 0:
        raise ValueError("No measures received - perhaps qrels and topics had no results in common")
    measures_sum = {}
    mean_dict = {}
    if measures is None:
        measures = list(next(iter(result.values())).keys())
    measures_remove = ["runid"]
    for m in measures_remove:
        if m in measures:
            measures.remove(m)
    measures_no_mean = set(["num_q", "num_rel", "num_ret", "num_rel_ret", "NumQ", "NumRel", "NumRet", "NumRelRet"])
    for val in result.values():
        for measure in measures:
            measure_val = val[measure]
            measures_sum[measure] = measures_sum.get(measure, 0.0) + measure_val
    if num_q is None:
        num_q = len(result.values())
    for measure, value in measures_sum.items():
        mean_dict[measure] = value / (1 if measure in measures_no_mean else num_q)
    return mean_dict


def _convert_measures(metrics : MEASURES_TYPE) -> Tuple[Sequence[Measure], Dict[Measure,str]]:
    from ir_measures import parse_trec_measure
    rtr = []
    rev_mapping = {}
    for m in metrics:
        if isinstance(m, Measure):
            rtr.append(m)
            continue
        elif isinstance(m, str):
            measures = parse_trec_measure(m)
            if len(measures) == 1:
                metric = measures[0]
                rtr.append(metric)
                rev_mapping[metric] = m
            elif len(measures) > 1:
                #m is family nickname, e.g. 'official;
                rtr.extend(measures)
            else:
                raise KeyError("Could not convert measure %s" % m)
        else:
            raise KeyError("Unknown measure %s of type %s" % (str(m), str(type(m))))
    assert len(rtr) > 0, "No measures were found in %s" % (str(metrics))
    return rtr, rev_mapping

  

class RenderFromPerQuery():
    """
    Responsible for calculating means of measures, applying significance testing, rounding and constructing dataframes
    """

    def __init__(self, systems, baseline=None, test_fn=None, correction=None, correction_alpha : float = 0.05, round=None, precompute_time=0):
        self.systems = systems
        self.baseline = baseline
        self.test_fn = test_fn
        self.correction = correction
        self.correction_alpha = correction_alpha
        self.round = round
        self.precompute_time = precompute_time
        # sysid -> qid -> measure -> value
        self.systemEvalDictsPerQ : Dict[int,Dict[str,Dict[str,float]]] = {}
        # sysid -> mrt
        self.mrts : Dict[int,float] = {}

    def add_metrics(self, sysid : int, evalRows : Dict[str, Dict[str, float]], mrt : float):
        if sysid >= len(self.systems):
            raise KeyError()
        self.systemEvalDictsPerQ[sysid] = evalRows
        self.mrts[sysid] = mrt
    
    @overload
    def averages(self, dataframe : Literal[True] = True, highlight : Optional[str] = None, mrt_needed : bool = False) -> pd.DataFrame: ...
    @overload
    def averages(self, dataframe : Literal[False], highlight : Optional[str] = None, mrt_needed : bool = False) -> Dict[str,Any]: ...

    def averages(self, dataframe : Union[Literal[True], Literal[False]] = True, highlight : Optional[str] = None, mrt_needed : bool = False) -> Union[Dict[str,Any], pd.DataFrame]:

        assert len(self.systemEvalDictsPerQ) == len(self.systems), "evaluation has not finished"
        baseline = self.baseline
        
        # this is needed for dataframe return
        evalMeasuresDicts = {}
        # one list of column values, one row for each system
        evalsRows = []
        # calculate means for all systems
        for i in range(len(self.systems)):
            evalMeasuresDict = _mean_of_measures(self.systemEvalDictsPerQ[i])
            if mrt_needed:
                evalMeasuresDict["mrt"] = self.precompute_time + self.mrts[i]
            actual_metric_names = list(evalMeasuresDict.keys())
            evalMeasures=[ self._apply_round(m, evalMeasuresDict[m]) for m in actual_metric_names]
            evalsRows.append([self.systems[i]]+evalMeasures) #
            evalMeasuresDicts[self.systems[i]] = evalMeasuresDict
        
        if not dataframe:
            return evalMeasuresDicts
        del(evalMeasuresDicts)

        highlight_cols = { m : "+"  for m in actual_metric_names }
        p_col_names : List[str] = []
        
        if baseline is not None:
            baselinePerQuery={}
            
            per_q_metrics = actual_metric_names.copy()
            if mrt_needed:
                per_q_metrics.remove("mrt")

            for m in per_q_metrics:
                baselinePerQuery[m] = np.array([ self.systemEvalDictsPerQ[baseline][q][m] for q in self.systemEvalDictsPerQ[baseline] ])

            for i in range(len(self.systems)):
                additionals: List[Optional[Union[float, int, complex]]] = []
                if i == baseline:
                    additionals = [None] * (3*len(per_q_metrics))
                else:
                    for m in per_q_metrics:
                        # we iterate through queries based on the baseline, in case run has different order
                        perQuery = np.array( [ self.systemEvalDictsPerQ[i][q][m] for q in self.systemEvalDictsPerQ[baseline] ])
                        delta_plus = (perQuery > baselinePerQuery[m]).sum()
                        delta_minus = (perQuery < baselinePerQuery[m]).sum()
                        p = self.test_fn(perQuery, baselinePerQuery[m])[1] # type: ignore[arg-type]
                        additionals.extend([delta_plus, delta_minus, p])
                evalsRows[i].extend(additionals)

            additional_col_names=[]
            for m in per_q_metrics:
                additional_col_names.append("%s +" % m)
                highlight_cols["%s +" % m] = "+"
                additional_col_names.append("%s -" % m)
                highlight_cols["%s -" % m] = "-"
                pcol = "%s p-value" % m
                additional_col_names.append(pcol)
                p_col_names.append(pcol)
            actual_metric_names.extend(additional_col_names)

        # its easier to build the dataframe, then apply the correction
        df = pd.DataFrame(evalsRows, columns=["name"] + actual_metric_names)
        
        # multiple testing correction. This adds two new columns for each measure experiencing statistical significance testing        
        if self.baseline is not None and self.correction is not None:
            import statsmodels.stats.multitest # type: ignore
            for pcol in p_col_names:
                pcol_reject = pcol.replace("p-value", "reject")
                pcol_corrected = pcol + " corrected"                
                reject, corrected, _, _ = statsmodels.stats.multitest.multipletests(df[pcol].drop(df.index[baseline]), alpha=self.correction_alpha, method=self.correction)
                insert_pos : int = df.columns.get_loc(pcol)
                # add reject/corrected values for the baseline
                reject = np.insert(reject, baseline, False)
                corrected = np.insert(corrected, baseline, np.nan)
                # add extra columns, put place directly after the p-value column
                df.insert(insert_pos+1, pcol_reject, reject)
                df.insert(insert_pos+2, pcol_corrected, corrected)

        if highlight == "color" or highlight == "colour" :
            df = df.style.apply(_color_cols, axis=0, col_type=highlight_cols) # type: ignore
        elif highlight == "bold":
            df = df.style.apply(_bold_cols, axis=0, col_type=highlight_cols) # type: ignore

        return df

    def _apply_round(self, measure, value):
        import builtins
        round = self.round
        if round is not None and isinstance(round, int):
            value = builtins.round(value, round)
        if round is not None and isinstance(round, dict) and measure in round:
            value = builtins.round(value, round[measure])
        return value
    
    @overload
    def perquery(self, dataframe: Literal[True] = True) -> pd.DataFrame: ...
    @overload
    def perquery(self, dataframe: Literal[False]) -> Dict[int,Dict[str, Dict[str,float]]]: ...

    def perquery(self, dataframe : Union[Literal[True], Literal[False]] = True) -> Union[Dict[int,Dict[str, Dict[str,float]]], pd.DataFrame]:
        """
        Return per-query results.
        DF has columns ``["name", "qid", "measure", "value"]``.
        """
        if not dataframe:
            return self.systemEvalDictsPerQ

        evalsRows = []
        for sysid in self.systemEvalDictsPerQ:
            evalMeasuresDict = self.systemEvalDictsPerQ[sysid]
            for qid in evalMeasuresDict:
                for measurename in evalMeasuresDict[qid]:
                    evalsRows.append([
                        self.systems[sysid], 
                        qid, 
                        measurename, 
                        self._apply_round(
                            measurename, 
                            evalMeasuresDict[qid][measurename]
                        )
                    ])
        
        perquery_df = pd.DataFrame(evalsRows, columns=["name", "qid", "measure", "value"]).sort_values(['name', 'qid'])
        if self.round is not None and isinstance(self.round, int):
            perquery_df["value"] = perquery_df["value"].round(self.round)
        return perquery_df




import pandas as pd
import numpy as np
import unittest
import pyterrier as pt
from pyterrier.terrier.stemmer import TerrierStemmer
from .base import TempDirTestCase
import tempfile
import shutil
import os

class TestScoring(TempDirTestCase):

    def test_scoring_tp(self):
        from pyterrier.terrier.retriever import TextScorer
        ts = TextScorer(wmodel="BM25", body_attr='text', background_index=pt.IndexFactory.of(self.here + "/fixtures/index"))
        score = ts([{'qid' : 'q1', 'query' : 'chemical reactions', 'docno' : 'd1', 'text' : 'professor proton mixed the chemicals'}])[0]['score']
        self.assertTrue(score != 0)

    def test_scoring_notp(self):
        from pyterrier.terrier.retriever import TextScorer
        ts = TextScorer(wmodel="Tf", body_attr='text', stemmer=None, stopwords=None)
        score = ts([{'qid' : 'q1', 'query' : 'chemical reactions', 'docno' : 'd1', 'text' : 'professor proton mixed the chemicals'}])[0]['score']
        self.assertEqual(score, 0.)
        score = ts([{'qid' : 'q1', 'query' : 'chemicals reaction', 'docno' : 'd1', 'text' : 'professor proton mixed the chemicals'}])[0]['score']
        self.assertEqual(score, 1.)

    def test_scoring_spanishtp(self):
        from pyterrier.terrier.retriever import TextScorer
        ts = TextScorer(wmodel="Tf", body_attr='text', stemmer='SpanishSnowballStemmer', stopwords=None)
        score = ts([{'qid' : 'q1', 'query' : 'checar', 'docno' : 'd1', 'text' : 'checa'}])[0]['score']
        self.assertEqual(score, 1.)

        ts = TextScorer(wmodel="Tf", body_attr='text', stemmer=TerrierStemmer.spanish, stopwords=None)
        score = ts([{'qid' : 'q1', 'query' : 'checar', 'docno' : 'd1', 'text' : 'checa'}])[0]['score']
        self.assertEqual(score, 1.)

    def test_scoring_text(self):
        pt.java.set_log_level("DEBUG")
        dataset = pt.get_dataset("vaswani")
        indexer = pt.TRECCollectionIndexer(
            self.test_dir, 
            meta= {'docno' : 26, 'body' : 2048},
            # The tags from which to save the text. ELSE is special tag name, which means anything not consumed by other tags.
            meta_tags = {'body' : 'ELSE'}
        )
        indexref = indexer.index(dataset.get_corpus())
        index = pt.IndexFactory.of(indexref)
        meta = index.getMetaIndex()
        self.assertTrue( "body" in meta.getKeys() )
        self.assertTrue( "compact memories have" in meta.getItem("body", 0) )
        print( meta.getItem("body", 1047))

        self._test_scoring_text(dataset, index, "org.terrier.python.TestModel$Constant")
        self._test_scoring_text(dataset, index, "Tf")
        self._test_scoring_text(dataset, index, "org.terrier.python.TestModel$TFOverN")
        self._test_scoring_text(dataset, index, "org.terrier.python.TestModel$F")
        self._test_scoring_text(dataset, index, "org.terrier.python.TestModel$Nt")
        self._test_scoring_text(dataset, index, "DPH")

    def _test_scoring_text(self, dataset, index, wmodel):
        
        br1 = pt.terrier.Retriever(index, wmodel=wmodel, metadata=["docno", "body"], num_results=5)
        input = dataset.get_topics().head(10)
        output1 = br1(input)
        self.assertTrue( "body" in output1.columns )
        input2 = output1[["qid", "query", "docno", "body"]]
        br2 = pt.terrier.TextScorer(background_index=index, wmodel=wmodel, verbose=True)
        output2 = br2(input2)
        self.assertTrue( "score" in output2.columns )

        joined = output1.merge(output2, on=["qid", "docno"])[["qid", "docno", "score_x", "score_y"]]
        print(joined)
        #TODO: there is a bug here. TextScorer should have the same score, but it doesnt; occasionally terms arent matched
        #self.assertTrue(np.array_equal(joined["score_x"].values, joined["score_y"].values))
        #self.assertEqual(pt.Evaluate(output1,dataset.get_qrels()), pt.Evaluate(output2,dataset.get_qrels()))
        

    def test_scoring_manual_empty(self):
        input = pd.DataFrame([["q1", "fox", "d1", ""]], columns=["qid", "query", "docno", "body"])
        from pyterrier.terrier.retriever import TextScorer
        scorer = TextScorer(wmodel="Tf")
        rtr = scorer(input)
        self.assertEqual(1, len(rtr))
        self.assertTrue("score" in rtr.columns)
        self.assertTrue("body" in rtr.columns)
        self.assertEqual(0, rtr.iloc[0]["score"])

    def test_scoring_manual(self):
        input = pd.DataFrame([["q1", "fox", "d1", "all the fox were fox"]], columns=["qid", "query", "docno", "body"])
        from pyterrier.terrier.retriever import TextScorer
        scorer = TextScorer(wmodel="Tf")
        rtr = scorer(input)
        self.assertEqual(1, len(rtr))
        self.assertTrue("score" in rtr.columns)
        self.assertTrue("body" in rtr.columns)
        self.assertEqual(2, rtr.iloc[0]["score"])

        scorer = TextScorer(wmodel="org.terrier.python.TestModel$TFOverN")
        rtr = scorer(input)
        self.assertEqual(1, len(rtr))
        self.assertTrue("score" in rtr.columns)
        self.assertEqual(2, rtr.iloc[0]["score"]) # tf / numdocs

    def test_scoring_manual_background(self):
        input = pd.DataFrame([["q1", "fox", "d1", "all the fox were fox"]], columns=["qid", "query", "docno", "body"])
        from pyterrier.terrier.retriever import TextScorer
        scorer = TextScorer(wmodel="Tf", background_index=pt.get_dataset("vaswani").get_index())
        rtr = scorer(input)
        self.assertEqual(1, len(rtr))
        self.assertTrue("score" in rtr.columns)
        self.assertEqual(2, rtr.iloc[0]["score"])

        index_background = pt.IndexFactory.of( pt.get_dataset("vaswani").get_index() )
        scorer = TextScorer(wmodel="org.terrier.python.TestModel$TFOverN", background_index=index_background)
        rtr = scorer(input)
        self.assertEqual(1, len(rtr))
        self.assertTrue("score" in rtr.columns)
        self.assertEqual(2/index_background.getCollectionStatistics().getNumberOfDocuments(), rtr.iloc[0]["score"])
        
    def test_scoring_qe(self):
        input = pd.DataFrame([["q1", "fox", "d1", "all the fox were fox"]], columns=["qid", "query", "docno", "body"])
        scorer = pt.terrier.retriever.TextIndexProcessor(pt.rewrite.Bo1QueryExpansion, takes="docs", returns="queries")
        rtr = scorer(input)
        self.assertTrue("qid" in rtr.columns)
        self.assertTrue("query" in rtr.columns)
        self.assertTrue("docno" not in rtr.columns)
        self.assertTrue("^" in rtr.iloc[0]["query"])



import unittest
import tempfile
import urllib
import pyterrier as pt
from .base import BaseTestCase


class TestArtifact(BaseTestCase):
    def test_from_hf_and_build_package(self):
        # this test covers a lot of stuff:
        #  - mapping of hf URLs (via from_hf)
        #  - downloading and validating package
        #  - loading correct artifact type
        #  - build_package
        for mem in [True, False]:
            try:
                index = pt.Artifact.from_hf('pyterrier/vaswani.terrier', memory=mem)
            except urllib.error.HTTPError as ex:
                if ex.code != 429: # too many requests ... can just ignore
                    raise
            retr = index.bm25(num_results=10)
            self.assertEqual(10, len(retr.search('chemical reactions')))
        with tempfile.TemporaryDirectory() as d:
            index.build_package(d+'/artifact.tar.lz4')


if __name__ == "__main__":
    unittest.main()



import unittest
import pyterrier as pt, pandas as pd
from .base import BaseTestCase

class TestInit(BaseTestCase):

    def test_set_property(self):
        pt.terrier.set_property("arbitrary.property", 40)
        self.assertEqual("40", pt.terrier.J.ApplicationSetup.appProperties.getProperty("arbitrary.property", "none"))



import pandas as pd
import pyterrier as pt
import os
import unittest
from .base import BaseTestCase, ensure_deprecated
import warnings

def parse_res_file(filename):
    results = []
    with open(filename, 'r') as file:
        for line in file:
            split_line = line.strip("\n").split(" ")
            results.append([split_line[0], split_line[2], float(split_line[4])])
    return results

def parse_query_result(filename):
    results = []
    with open(filename, 'rt') as file:
        for line in file:
            split_line = line.strip("\n").split(" ")
            results.append([split_line[1], float(split_line[2])])
    return results

class TestBatchRetrieve(BaseTestCase):

    def test_candidate_set_one_doc(self):
        if not pt.terrier.check_version("5.3"):
            self.skipTest("Requires Terrier 5.3")
        indexloc = self.here + "/fixtures/index/data.properties"
        # docid 50 == docno 51
        input_set = pd.DataFrame([["q1", "light", 50]], columns=["qid", "query", "docid"])
        retr = pt.terrier.Retriever(indexloc)
        
        # this test the implementation of __call__() redirecting to transform()
        for result in [retr.transform(input_set), retr(input_set)]:
            result = retr.transform(input_set)
            self.assertTrue("qid" in result.columns)
            self.assertTrue("docno" in result.columns)
            self.assertTrue("score" in result.columns)
            self.assertTrue("rank" in result.columns)
            self.assertEqual(1, len(result))
            row = result.iloc[0]
            self.assertEqual("q1", row["qid"])
            self.assertEqual("51", row["docno"])
            self.assertEqual(pt.model.FIRST_RANK, row["rank"])
            self.assertTrue(row["score"] > 0)

    def test_br_cutoff(self):
        indexloc = self.here + "/fixtures/index/data.properties"
        
        input_set = pd.DataFrame([
                    ["q1", "chemical"],
                ],
            columns=["qid", "query"])
        retr = pt.terrier.Retriever(indexloc) % 10
        result = retr.transform(input_set)
        self.assertEqual(10, len(result))

    def test_br_query_toks(self):
        from pandas.testing import assert_frame_equal
        indexloc = self.here + "/fixtures/index/data.properties"
        
        retr = pt.terrier.Retriever(indexloc)

        # check that terrier doesnt error about queries with punctation
        res1 = retr.search("What are chemical reactions?")
        res2 = retr.search("chemical reactions") # 'what' and 'are' are stopwords
        assert_frame_equal(res1[["qid", "docno", "score", "rank"]], res2[["qid", "docno", "score", "rank"]])

        query_terrier = 'applypipeline:off chemic^2 reaction^0.5'
        result_terrier = retr.search(query_terrier)

        query_matchop = '#combine:0=2:1=0.5(chemic reaction)'
        result_matchop = retr.search(query_matchop)

        query_toks = { 'chemic' : 2, 'reaction' : 0.5}
        result_toks = retr.transform(pd.DataFrame([['1', query_toks]], columns=['qid', 'query_toks']))
        
        self.assertEqual(len(result_terrier), len(result_matchop))
        self.assertEqual(len(result_terrier), len(result_toks))
        assert_frame_equal(result_terrier[["qid", "docno", "score", "rank"]], result_matchop[["qid", "docno", "score", "rank"]])
        assert_frame_equal(result_terrier[["qid", "docno", "score", "rank"]], result_toks[["qid", "docno", "score", "rank"]])

    def test_br_cutoff_stability(self):
        indexloc = self.here + "/fixtures/index/data.properties"
        input_set = pd.DataFrame([
                    ["q1", "chemical"],
                ],
            columns=["qid", "query"])
        br_cut_3 = pt.terrier.Retriever(indexloc, wmodel='Tf') % 3
        br_3 = pt.terrier.Retriever(indexloc, wmodel='Tf', num_results=3)
        #br = pt.terrier.Retriever(indexloc, wmodel='Tf')
        #print(br.transform(input_set))

        result_cut = br_cut_3.transform(input_set)
        result_tr = br_3.transform(input_set)
        print("Rank cutoff operator")
        print(result_cut.docno)
        print("terrier cutoff")
        print(result_tr.docno)        
        pd.testing.assert_series_equal(result_cut.docno, result_tr.docno)
    
    def test_br_col_passthrough(self):
        indexloc = self.here + "/fixtures/index/data.properties"
        
        input_set = pd.DataFrame([
                    ["q1", "chemical", "u1"],
                ],
            columns=["qid", "query", "username"])
        retr = pt.terrier.Retriever(indexloc) % 10
        result = retr.transform(input_set)
        self.assertIn("username", result.columns)

    def test_br_mem(self):
        indexloc = self.here + "/fixtures/index/data.properties"
        memindex = pt.IndexFactory.of(indexloc, memory=True)
        pindex = pt.java.cast("org.terrier.structures.IndexOnDisk", memindex)
        self.assertEqual("fileinmem", pindex.getIndexProperty("index.lexicon.data-source", "notfound"))
        retr = pt.terrier.Retriever(memindex)
        retr.search("chemical reactions")

    def test_br_empty(self):
        indexloc = self.here + "/fixtures/index/data.properties"
        
        input_set = pd.DataFrame([
                    ["q1", ""],
                ],
            columns=["qid", "query"])
        retr = pt.terrier.Retriever(indexloc)
        with warnings.catch_warnings(record=True) as w:
            result = retr.transform(input_set)
            assert "Skipping empty query" in str(w[-1].message)
        self.assertEqual(0, len(result))

    def test_candidate_set_two_doc(self):
        if not pt.terrier.check_version("5.3"):
            self.skipTest("Requires Terrier 5.3")

        indexloc = self.here + "/fixtures/index/data.properties"
        # docid 50 == docno 51
        # docid 66 == docno 67

        input_set = pd.DataFrame([
                    ["q1", "light", 50],
                    ["q1", None, 66]
                ],
            columns=["qid", "query", "docid"])
        retr = pt.terrier.Retriever(indexloc)
        result = retr.transform(input_set)
        self.assertTrue("qid" in result.columns)
        self.assertTrue("docno" in result.columns)
        self.assertTrue("score" in result.columns)
        self.assertEqual(2, len(result))

    # this also tests different index-like inputs, namely:
    # a String index location, an IndexRef, and an Index
    def test_one_term_query_correct_docid_score_rank(self):

        indexloc = self.here + "/fixtures/index/data.properties"
        jindexref = pt.IndexRef.of(indexloc)
        jindex = pt.IndexFactory.of(jindexref)
        for indexSrc in (indexloc, jindexref, jindex):
            retr = pt.terrier.Retriever(indexSrc)
            result = retr.search("light")
            exp_result = parse_query_result(os.path.dirname(
                os.path.realpath(__file__)) + "/fixtures/light_results")
            i=0
            for index, row in result.iterrows():
                self.assertEqual(row['docno'], exp_result[index][0])
                self.assertAlmostEqual(row['score'], exp_result[index][1])
                self.assertEqual(pt.model.FIRST_RANK + i, row["rank"])
                i+=1

        jindex.close()

    def test_two_term_query_correct_qid_docid_score(self):
        JIR = pt.java.autoclass('org.terrier.querying.IndexRef')
        indexref = JIR.of(self.here + "/fixtures/index/data.properties")
        retr = pt.terrier.Retriever(indexref)
        input = pd.DataFrame([["1", "Stability"], ["2", "Generator"]], columns=['qid', 'query'])
        result = retr.transform(input)
        exp_result = parse_res_file(os.path.dirname(
            os.path.realpath(__file__)) + "/fixtures/two_queries_result")
        for index, row in result.iterrows():
            self.assertEqual(row['qid'], exp_result[index][0])
            self.assertEqual(row['docno'], exp_result[index][1])
            self.assertAlmostEqual(row['score'], exp_result[index][2])

        input = pd.DataFrame([[1, "Stability"], [2, "Generator"]], columns=['qid', 'query'])
        result = retr.transform(input)
        exp_result = parse_res_file(os.path.dirname(
            os.path.realpath(__file__)) + "/fixtures/two_queries_result")
        for index, row in result.iterrows():
            self.assertEqual(str(row['qid']), exp_result[index][0])
            self.assertEqual(row['docno'], exp_result[index][1])
            self.assertAlmostEqual(row['score'], exp_result[index][2])

    def test_num_results(self):
        JIR = pt.java.autoclass('org.terrier.querying.IndexRef')
        indexref = JIR.of(self.here+"/fixtures/index/data.properties")
        retr = pt.terrier.Retriever(indexref, num_results=10)
        input=pd.DataFrame([["1", "Stability"]],columns=['qid','query'])
        result = retr.transform(input)
        self.assertEqual(len(result), 10)

        if not pt.terrier.check_version("5.5"):
            return

        retr = pt.terrier.Retriever(indexref, num_results=1001)        
        result = retr.search("results")
        self.assertEqual(len(result), 1001)

    def test_num_manual_wmodel(self):
        JIR = pt.java.autoclass('org.terrier.querying.IndexRef')
        Tf = pt.java.autoclass("org.terrier.matching.models.Tf")()
        indexref = JIR.of(self.here+"/fixtures/index/data.properties")
        from jnius import JavaException
        try:
            retr = pt.terrier.Retriever(indexref, wmodel=Tf)
            input=pd.DataFrame([["1", "Stability"]],columns=['qid','query'])
            result = retr.transform(input)
        except JavaException as ja:
            print(ja.stacktrace)
            raise ja

    def test_num_python_wmodel(self):
        indexref = self.here+"/fixtures/index/data.properties"
        Tf = lambda keyFreq, posting, entryStats, collStats: posting.getFrequency()
        retr = pt.terrier.Retriever(indexref, wmodel=Tf)
        input=pd.DataFrame([["1", "Stability"]],columns=['qid','query'])
        result = retr.transform(input)

    def test_threading_manualref(self):
        
        if not pt.terrier.check_version("5.5"):
            self.skipTest("Requires Terrier 5.5")

        topics = pt.get_dataset("vaswani").get_topics().head(8)

        #this test ensures that we operate when the indexref is specified to be concurrent 
        JIR = pt.java.autoclass('org.terrier.querying.IndexRef')
        indexref = JIR.of("concurrent:" + self.here+"/fixtures/index/data.properties")
        retr = pt.terrier.Retriever(indexref, threads=4)
        result = retr.transform(topics)

        #check that use of a callback model works under threading
        Tf = lambda keyFreq, posting, entryStats, collStats: posting.getFrequency()
        retr = pt.terrier.Retriever(indexref, threads=4, wmodel=Tf)
        result = retr.transform(topics)

    def test_threading_selfupgrade(self):
        if not pt.terrier.check_version("5.5"):
            self.skipTest("Requires Terrier 5.5")

        topics = pt.get_dataset("vaswani").get_topics().head(10)

        #this test ensures we can upgrade the indexref to be concurrent
        JIR = pt.java.autoclass('org.terrier.querying.IndexRef')
        indexref = JIR.of(self.here+"/fixtures/index/data.properties")
        retr = pt.terrier.Retriever(indexref, threads=5)
        result = retr.transform(topics)

    def test_terrier_retrieve_alias(self):
        # based off test_candidate_set_one_doc
        if not pt.terrier.check_version("5.3"):
            self.skipTest("Requires Terrier 5.3")
        indexloc = self.here + "/fixtures/index/data.properties"
        # docid 50 == docno 51
        input_set = pd.DataFrame([["q1", "light", 50]], columns=["qid", "query", "docid"])

        @ensure_deprecated
        def do_test(name, retr):
            # delay instantiation to here
            retr = retr(indexloc)
            self.subTest(name)
            # this test the implementation of __call__() redirecting to transform()
            for result in [retr.transform(input_set), retr(input_set)]:
                result = retr.transform(input_set)
                self.assertTrue("qid" in result.columns)
                self.assertTrue("docno" in result.columns)
                self.assertTrue("score" in result.columns)
                self.assertTrue("rank" in result.columns)
                self.assertEqual(1, len(result))
                row = result.iloc[0]
                self.assertEqual("q1", row["qid"])
                self.assertEqual("51", row["docno"])
                self.assertEqual(pt.model.FIRST_RANK, row["rank"])
                self.assertTrue(row["score"] > 0)
        
        for name, retr in [ ("TerrierRetrieve", pt.TerrierRetrieve), ("BatchRetrieve", pt.BatchRetrieve) ]:
            do_test(name, retr)


class TestTerrierRetrieverBasic(pt.testing.TransformerTestCase):
    @staticmethod
    def get_transformer():
        index = pt.terrier.TerrierIndex.load(os.path.dirname(os.path.realpath(__file__)) + '/fixtures/index/')
        return index.bm25()


if __name__ == "__main__":
    unittest.main()



import unittest
import pyterrier as pt
import tempfile
from .base import TempDirTestCase

class TestDunder(TempDirTestCase):

    def test_callable_wmodel_dunders(self):
        testPosting = pt.java.autoclass("org.terrier.structures.postings.BasicPostingImpl")(0,1)

        lambdafn = lambda keyFreq, posting, entryStats, collStats: posting.getFrequency()
        callback, wmodel = pt.terrier.retriever._function2wmodel(lambdafn)
        
        byterep = pt.java.bytebuffer_to_array(wmodel.scoringClass.serializeFn())
        import dill as pickle
        from dill import extend
        #see https://github.com/SeldonIO/alibi/issues/447#issuecomment-881552005
        extend(use_dill=False)    
        fn = pickle.loads(byterep)
        self.assertEqual(
            lambdafn(1, testPosting, None, None),
            fn(1, testPosting, None, None),
            )

        wmodel.__getstate__()
        rtr = wmodel.__reduce__()
        
        #check the byte array is picklable
        pickle.dumps(rtr[1][0])
        #check object is picklable
        pickle.dumps(wmodel)
        #check can be unpickled too
        wmodel2 = pickle.loads(pickle.dumps(wmodel))

        score1 = wmodel.score(testPosting)
        score2 = wmodel2.score(testPosting)
        self.assertEqual(score1, score2)

        #check newly unpickled can still be pickled
        pickle.dumps(wmodel2)
        wmodel3 = pickle.loads(pickle.dumps(wmodel2))
        score3 = wmodel3.score(testPosting)
        self.assertEqual(score1, score3)


    def test_wmodel_dunders(self):

        wmodel = pt.java.autoclass("org.terrier.matching.models.BM25")()
        wmodel.__reduce__()
        wmodel.__getstate__()
        rtr = wmodel.__reduce__()
        pt.java.cast("org.terrier.matching.models.BM25", rtr[0](*rtr[1]))
        import pickle
        #import dill as pickle
        #check the byte array is picklable
        print(rtr[1][0])
        pickle.dumps(rtr[1][0])
        pickle.dumps(wmodel)

    def test_index_dunders(self):
        indexref = pt.datasets.get_dataset("vaswani").get_index()
        i1 = pt.IndexFactory.of(indexref)
        i2 = pt.IndexFactory.of(indexref)
        i12 = i1 + i2
        self.assertIsNotNone(i12)
        self.assertEqual(
            i12.getCollectionStatistics().getNumberOfDocuments(), 
            i1.getCollectionStatistics().getNumberOfDocuments()
            + i2.getCollectionStatistics().getNumberOfDocuments())
        self.assertEqual( len(i1), i1.getCollectionStatistics().getNumberOfDocuments() )
        self.assertEqual( len(i12), len(i1) + len(i2) )
            
        self.assertTrue(i12.hasIndexStructure("inverted"))
        self.assertTrue(i12.hasIndexStructure("lexicon"))
        self.assertTrue(i12.hasIndexStructure("document"))
        self.assertTrue(i12.hasIndexStructure("meta"))

    def test_dunders(self):
        import pandas as pd
        df = pd.DataFrame({
            'docno':
                ['1', '2', '3'],
            'text':
                ['He ran out of money, so he had to stop playing',
                 'The waves were crashing on the shore; it was a',
                 'The body may perhaps compensates for the loss']
        })
        import pyterrier as pt
        pd_indexer = pt.IterDictIndexer(self.test_dir, stopwords=pt.TerrierStopwords.none, stemmer=pt.TerrierStemmer.none)
        indexref = pd_indexer.index(df.to_dict(orient='records'))
        index = pt.IndexFactory.of(indexref)
        self.assertIsNotNone(index)
        self.assertIsNotNone(index.getLexicon())
        self.assertTrue("__getitem__" in dir(index.getLexicon()))
        crashingSeen = False

        # test out __len__ mapping
        self.assertEqual(len(index.getLexicon()), index.getLexicon().numberOfEntries())
        # lexicon is Iterable, test the Iterable mapping of jnius
        for t in index.getLexicon():
            if t.getKey() == "crashing":
                crashingSeen = True
                break
        self.assertTrue(crashingSeen)
        # test our own __getitem__ mapping
        self.assertTrue("crashing" in index.getLexicon())
        self.assertEqual(1, index.getLexicon()["crashing"].getFrequency())
        self.assertFalse("dentish" in index.getLexicon())
        
        # test Map$Entry can be decoded like a tuple
        # both within an iterator, and separately
        x, y = index.getLexicon().getLexiconEntry(0)
        term_count = 0
        for term, entry in index.getLexicon():
            term_count += 1
        self.assertEqual(term_count, index.getLexicon().numberOfEntries())
        term_count = 0
        for lee in index.getLexicon():
            term_count += 1
            lee.getKey()
            lee.getValue()
        self.assertEqual(term_count, index.getLexicon().numberOfEntries())

        # now test that IterablePosting has had its dunder methods added
        postings = index.getInvertedIndex().getPostings(index.getLexicon()["crashing"])
        count = 0
        for p in postings:
            count += 1
            self.assertEqual(1, p.getId())
            self.assertEqual(1, p.getFrequency())
        self.assertEqual(1, count)

if __name__ == "__main__":
    unittest.main()



import pandas as pd
from typing import Union, List, Literal
from typing import Iterable
import pyterrier as pt


class IRDSDatasetProvider(pt.datasets.DatasetProvider):
    def get_dataset(self, name) -> pt.datasets.Dataset:
        return IRDSDataset(name)

    def list_dataset_names(self) -> Iterable[str]:
        import ir_datasets
        return list(ir_datasets.registry)


class IRDSDataset(pt.datasets.Dataset):
    def __init__(self, irds_id):
        import ir_datasets
        self._irds_id = irds_id
        self._irds_ref = ir_datasets.load(self._irds_id)

    def irds_ref(self):
        return self._irds_ref

    def get_corpus_iter(self, verbose=True, start=0, count=None):
        ds = self.irds_ref()
        if not ds.has_docs():
            raise NotImplementedError(f"{self!r} doesn't support get_corpus_iter")
        it = ds.docs_iter()
        total = ds.docs_count()

        # use slicing if requested
        if start > 0 or count is not None:
            if count is not None:
                it = it[start:start+count]
                total = count
            else:
                it = it[start:]
                total -= start

        # tqdm support
        if verbose:
            it = pt.tqdm(it, desc=f'{self._irds_id} documents', total=total)

        # rewrite to follow pyterrier std
        def gen():
            for doc in it:
                doc = doc._asdict()
                # pyterrier uses "docno"
                doc['docno'] = doc.pop('doc_id')
                yield doc

        # provide a __len__ (e.g., so tqdm shows progress)
        return pt.utils.GeneratorLen(gen(), total)

    def get_corpus_lang(self):
        ds = self.irds_ref()
        if not ds.has_docs():
            return None
        return ds.docs_lang()

    def get_topics(self, variant=None):
        """
            Returns the topics, as a dataframe, ready for retrieval. 
        """
        ds = self.irds_ref()
        if not ds.has_queries():
            raise NotImplementedError(f"{self._irds_id} doesn't support get_topics")
        qcls = ds.queries_cls()
        assert variant is None or variant in qcls._fields[1:], f"{self._irds_id} only supports the following topic variants {qcls._fields[1:]}"
        df = pd.DataFrame(ds.queries_iter())

        df.rename(columns={"query_id": "qid"}, inplace=True) # pyterrier uses "qid"

        if variant is not None:
            # Some datasets have a query field called "query". We need to remove it or
            # we'll end up with multiple "query" columns, which will cause problems
            # because many components are written assuming no columns have the same name.
            if variant != 'query' and 'query' in df.columns:
                df.drop(columns=['query'], axis=1, inplace=True)
            df.rename(columns={variant: "query"}, inplace=True) # user specified which version of the query they want
            df.drop(columns=df.columns.difference(['qid','query']), axis=1, inplace=True)
        elif len(qcls._fields) == 2:
            # auto-rename single query field to "query" if there's only query_id and that field
            df.rename(columns={qcls._fields[1]: "query"}, inplace=True)
        else:
            print(f'There are multiple query fields available: {qcls._fields[1:]}. To use with pyterrier, provide variant or modify dataframe to add query column.')
        return df

    def get_topics_lang(self):
        ds = self.irds_ref()
        if not ds.has_queries():
            return None
        return ds.queries_lang()

    def get_qrels(self, variant=None):
        """ 
            Returns the qrels, as a dataframe, ready for evaluation.
        """
        ds = self.irds_ref()
        if not ds.has_qrels():
            raise NotImplementedError(f"{self._irds_id} doesn't support get_qrels")
        qrelcls = ds.qrels_cls()
        qrel_fields = [f for f in qrelcls._fields if f not in ('query_id', 'doc_id', 'iteration')]
        assert variant is None or variant in qrel_fields, f"{self._irds_id} only supports the following qrel variants {qrel_fields}"
        df = pd.DataFrame(ds.qrels_iter())

        # pyterrier uses "qid" and "docno"
        df.rename(columns={
            "query_id": "qid",
            "doc_id": "docno"}, inplace=True)

        # pyterrier uses "label"
        if variant is not None:
            df.rename(columns={variant: "label"}, inplace=True)
        if len(qrel_fields) == 1:
            # usually "relevance"
            df.rename(columns={qrel_fields[0]: "label"}, inplace=True)
        elif 'relevance' in qrel_fields:
            print(f'There are multiple qrel fields available: {qrel_fields}. Defaulting to "relevance", but to use a different one, supply variant')
            df.rename(columns={'relevance': "label"}, inplace=True)
        else:
            print(f'There are multiple qrel fields available: {qrel_fields}. To use with pyterrier, provide variant or modify dataframe to add query column.')

        return df

    def get_results(self, variant=None) -> pd.DataFrame:
        """ 
            Returns a standard result set provided by the dataset. This is useful for re-ranking experiments.
        """
        ds = self.irds_ref()
        if not ds.has_scoreddocs():
            raise NotImplementedError(f"{self._irds_id} doesn't support get_results")
        result = pd.DataFrame(ds.scoreddocs)
        result = result.rename(columns={'query_id': 'qid', 'doc_id': 'docno'}) # convert irds field names to pyterrier names
        result.sort_values(by=['qid', 'score', 'docno'], ascending=[True, False, True], inplace=True) # ensure data is sorted by qid, -score, did
        # result doesn't yet contain queries (only qids) so load and merge them in
        topics = self.get_topics(variant)
        result = pd.merge(result, topics, how='left', on='qid')
        return result

    def _describe_component(self, component):
        ds = self.irds_ref()
        if component == "topics":
            if ds.has_queries():
                fields = ds.queries_cls()._fields[1:]
                if len(fields) > 1:
                    return list(fields)
                return True
            return None
        if component == "qrels":
            if ds.has_qrels():
                fields = [f for f in ds.qrels_cls()._fields if f not in ('query_id', 'doc_id', 'iteration')]
                if len(fields) > 1:
                    return list(fields)
                return True
            return None
        if component == "corpus":
            return ds.has_docs() or None
        if component == "results":
            return ds.has_scoreddocs() or None
        return None

    def info_url(self):
        top_id = self._irds_id.split('/', 1)[0]
        suffix = f'#{self._irds_id}' if top_id != self._irds_id else ''
        return f'https://ir-datasets.com/{top_id}.html{suffix}'

    def __repr__(self):
        return f"IRDSDataset({self._irds_id!r})"

    def text_loader(
        self,
        fields: Union[List[str], str, Literal['*']] = '*',
        *,
        verbose: bool = False,
    ) -> pt.Transformer:
        """Create a transformer that loads text fields from an ir_datasets dataset into a DataFrame by docno.

        Args:
            fields: The fields to load from the dataset. If '*', all fields will be loaded.
            verbose: Whether to print debug information.
        """
        return IRDSTextLoader(self, fields, verbose=verbose)


class IRDSTextLoader(pt.Transformer):
    """A transformer that loads text fields from an ir_datasets dataset into a DataFrame by docno."""
    schematic = {'label': 'TextLoader'}
    def __init__(
        self,
        dataset: IRDSDataset,
        fields: Union[List[str], str, Literal['*']] = '*',
        *,
        verbose=False
    ):
        """Initialise the transformer with the index to load metadata from.

        Args:
            dataset: The dataset to load text from.
            fields: The fields to load from the dataset. If '*', all fields will be loaded.
            verbose: Whether to print debug information.
        """
        if not dataset.irds_ref().has_docs():
            raise ValueError(f"Dataset {dataset} does not provide docs")
        docs_cls = dataset.irds_ref().docs_cls()

        available_fields = [f for f in docs_cls._fields if f != 'doc_id' and docs_cls.__annotations__[f] is str]
        if fields == '*':
            fields = available_fields
        else:
            if isinstance(fields, str):
                fields = [fields]
            missing_fields = set(fields) - set(available_fields)
            if missing_fields:
                raise ValueError(f"Dataset {dataset} did not have requested metaindex keys {list(missing_fields)}. "
                                 f"Keys present in metaindex are {available_fields}")

        self.dataset = dataset
        self.fields = fields
        self.verbose = verbose

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        """Load text fields from the dataset into the input DataFrame.

        Args:
            inp: The input DataFrame. Must contain 'docno'.

        Returns:
            A new DataFrame with the text columns appended.
        """
        if 'docno' not in inp.columns:
            raise ValueError(f"input missing 'docno' column, available columns: {list(inp.columns)}")
        irds = self.dataset.irds_ref()
        docstore = irds.docs_store()
        docnos = inp.docno.values.tolist()

        # Load the new data
        fields = ['doc_id'] + self.fields
        set_docnos = set(docnos)
        it = (tuple(getattr(doc, f) for f in fields) for doc in docstore.get_many_iter(set_docnos))
        if self.verbose:
            it = pt.tqdm(it, unit='d', total=len(set_docnos), desc='IRDSTextLoader') # type: ignore
        metadata = pd.DataFrame(list(it), columns=fields).set_index('doc_id')
        metadata_frame = metadata.loc[docnos].reset_index(drop=True)

        # append the input and metadata frames
        inp = inp.drop(columns=self.fields, errors='ignore') # make sure we don't end up with duplicates
        inp = inp.reset_index(drop=True) # reset the index to default (matching metadata_frame)
        return pd.concat([inp, metadata_frame], axis='columns')

    def fuse_rank_cutoff(self, k):
        return pt.RankCutoff(k) >> self



import ast
from typing import Generator, Tuple, Type, Any, List, Optional
import pyterrier_flake8_ext


class JavaCheck:
    """McCabe cyclomatic complexity checker."""
    name = 'javacheck'
    version = pyterrier_flake8_ext.__version__
    _missing_anno_tmpl = "PT100 {} uses java but is not decorated with @pt.java.required"
    _extra_anno_tmpl = "PT101 {} is decorated with @pt.java.required but doesn't use java"
    _required_order = "PT102 {} is decorated with @pt.java.required but isn't the last decorator"

    def __init__(self, tree: ast.AST):
        self.tree = tree

    def run(self) -> Generator[Tuple[int, int, str, Type[Any]], None, None]:
        yield from self._run(self.tree)

    def _run(self, node: ast.AST, context: List[str] = []) -> Generator[Tuple[int, int, str, Type[Any]], None, None]:
        if isinstance(node, ast.FunctionDef) or isinstance(node, ast.ClassDef):
            has_java_anno, hard_require = self._has_java_required(node)
            if has_java_anno:
                err = self._required_order_error(node)
                if err:
                    yield err
            java_usage = None
            if isinstance(node, ast.FunctionDef):
                for child in node.body:
                    java_usage = self._uses_java(child)
                    if java_usage: break
            if isinstance(node, ast.ClassDef):
                for child in node.body:
                    if isinstance(child, ast.FunctionDef):
                        for child2 in child.body:
                            java_usage = self._uses_java(child2)
                            if java_usage: break
                    else:
                        java_usage = self._uses_java(child)
                    if java_usage: break
            if not has_java_anno and java_usage and isinstance(node, ast.FunctionDef):
                text = self._missing_anno_tmpl.format('.'.join(context + [node.name]))
                yield java_usage.lineno, java_usage.col_offset, text, type(self)
            elif has_java_anno and not java_usage and hard_require:
                text = self._extra_anno_tmpl.format('.'.join(context + [node.name]))
                yield node.lineno, node.col_offset, text, type(self)
            if has_java_anno and isinstance(node, ast.ClassDef):
                # simulate @required annotations on the methods defined by this class
                for child in node.body:
                    if isinstance(child, ast.FunctionDef):
                        child.decorator_list.append(ast.Name('__required__soft'))

        if isinstance(node, ast.FunctionDef):
            context = context + [node.name]

        if isinstance(node, ast.ClassDef):
            context = context + [node.name]

        # we probably don't need such an extensive list when searching for functions/classes
        for attr in ['body', 'value', 'values', 'target', 'targets', 'left', 'right', 'elt', 'elts', 'keys', 'values', 'test', 'orelse', 'generators', 'func', 'args', 'keywords', 'format_spec']:
            if hasattr(node, attr):
                value = getattr(node, attr)
                if value is not None:
                    if hasattr(value, '__len__'):
                        for child in value:
                            yield from self._run(child, context)
                    else:
                        yield from self._run(value, context)

    def _has_java_required(self, function: ast.FunctionDef) -> Tuple[bool, bool]:
        for decorator in function.decorator_list:
            if isinstance(decorator, ast.Name):
                if decorator.id in ('required', 'required_raise'):
                    return True, True
                if decorator.id == '__required__soft':
                    return True, False
            elif isinstance(decorator, ast.Attribute):
                if decorator.attr in ('required', 'required_raise'):
                    return True, True
                if decorator.attr == '__required__soft':
                    return True, False
        return False, False

    def _required_order_error(self, function: ast.FunctionDef) -> Optional[Tuple[int, int, str, Type[Any]]]:
        for i, decorator in enumerate(function.decorator_list):
            is_last = i == len(function.decorator_list) - 1
            if isinstance(decorator, ast.Name):
                if decorator.id in ('required', 'required_raise') and not is_last:
                    text = self._required_order.format(function.name)
                    return decorator.lineno, decorator.col_offset, text, type(self)
            elif isinstance(decorator, ast.Attribute) and not is_last:
                if decorator.attr in ('required', 'required_raise'):
                    text = self._required_order.format(function.name)
                    return decorator.lineno, decorator.col_offset, text, type(self)
        return None

    def _uses_java(self, node: ast.AST) -> Optional[ast.AST]:
        if isinstance(node, ast.Name):
            if node.id in ('autoclass', 'cast', 'J'):
                return node
        elif isinstance(node, ast.Attribute):
            if node.attr in ('autoclass', 'cast', 'J'):
                return node
        elif isinstance(node, ast.Import):
            for name in node.names:
                if name.name == 'jnius':
                    return node
        elif isinstance(node, ast.ImportFrom):
            if node.module == 'jnius':
                return node
        elif isinstance(node, ast.ClassDef):
            return None # break search at class definintions
        elif isinstance(node, ast.FunctionDef):
            return None # break search at function definintions

        for attr in ['body', 'value', 'values', 'target', 'targets', 'left', 'right', 'elt', 'elts', 'keys', 'values', 'test', 'orelse', 'generators', 'func', 'args', 'keywords', 'format_spec']:
            if hasattr(node, attr):
                value = getattr(node, attr)
                if value is not None:
                    if hasattr(value, '__len__'):
                        for child in value:
                            res = self._uses_java(child)
                            if res:
                                return res
                    else:
                        res = self._uses_java(value)
                        if res:
                            return res
        return None



import re
import math
import itertools
import numpy as np
import pandas as pd
from typing import Any, Dict, Iterable, List, Sequence, Set, Optional, Union, overload

IterDictRecord = Dict[str, Any]
IterDict = Iterable[IterDictRecord]

# This file has useful methods for using the Pyterrier Pandas datamodel

# the first rank SHOULD be 0, see the standard "Welcome to TREC email"
FIRST_RANK = 0

# set to True to ensure that the resulting dataframe is correctly /ordered/
# as well as having correct ranks assigned
STRICT_SORT = False

def add_ranks(df : pd.DataFrame, single_query=False) -> pd.DataFrame:
    """
        Canonical method for adding a rank column which is calculated based on the score attribute
        for each query. Note that the dataframe is NOT sorted by this operation (this is defined by STRICT_SORT).
        The dataframe is modified, i.e. inplace.

        Arguments
            df: dataframe to create rank attribute for
            single_query (bool): whether the dataframe contains only a single-query or not. This method will be quicker for single-query dataframes
    """
    if "score" not in df.columns:
        raise KeyError("Must have score column to add ranks to dataframe. Found columns were %s" % str(df.columns.values.tolist()))
    df.drop(columns=["rank"], errors="ignore", inplace=True)
    if len(df) == 0:
        df["rank"] = pd.Series(index=df.index, dtype='int64')
        return df

    # remove group when single_query is True
    if single_query:
        # -1 assures that first rank will be FIRST_RANK
        df["rank"] = df["score"].rank(ascending=False, method="first").astype(int) -1 + FIRST_RANK
        if STRICT_SORT:
            df.sort_values(["rank"], ascending=True, inplace=True)
        return df

    # -1 assures that first rank will be FIRST_RANK
    df["rank"] = df.groupby("qid", sort=False)["score"].rank(ascending=False, method="first").astype(int) -1 + FIRST_RANK
    if STRICT_SORT:
        df.sort_values(["qid", "rank"], ascending=[True,True], inplace=True)
    return df

def document_columns(df : pd.DataFrame) -> Sequence[str]:
    """
        Given a dataframe, returns the names of all columns that contain attributes that are 
        concerned with a document, or the relationship between a document and a query.

        It is defined as the complement of query_columns().
    """
    return list(df.columns.difference(query_columns(df, qid=False)))

def query_columns(df : pd.DataFrame, qid=True) -> Sequence[str]:
    """
        Given a dataframe, returns the names of all columns that contain the current query or
        previous generations of the query (as performed by `push_queries()`). 

        Any saved_docs_0 column is also included.

        Arguments:
            df: Dataframe of queries to consider
            qid: whether to include the "qid" column in the returned list
    """
    columns=set(df.columns)
    rtr = []
    if qid and "qid" in columns:
        rtr.append("qid")
    if "query" in columns:
        rtr.append("query")
    for c in columns:
        if c.startswith("q") and c not in rtr:
            if c == 'qid' and not qid:
                continue
            rtr.append(c)
    import re
    saved_docs_col_re = re.compile('^stashed_results_[\\d]+')
    for c in columns:
        if saved_docs_col_re.search(c):
            rtr.append(c)
    return rtr


def push_queries(df: pd.DataFrame, *, keep_original: bool = False, inplace: bool = False) -> pd.DataFrame:
    """
        Changes a dataframe such that the "query" column becomes "query_0", and any
        "query_0" columns becames "query_1" etc.

        Arguments:
            df: Dataframe with a "query" column
            keep_original: if True, the query column is also left unchanged. Useful for client code. 
                Defaults to False.
            inplace: if False, a copy of the dataframe is returned. If True, changes are made to the
                supplied dataframe. Defaults to False. 
    """
    cols = set(df.columns)
    if "query" not in cols:
        raise KeyError(f"Expected a query column, but found {list(cols)}")
    if not inplace:
        df = df.copy()
    prev_col = 'query'
    rename_cols = {}
    for query_idx in itertools.count():
        next_col = f'query_{query_idx}'
        if prev_col in cols:
            rename_cols[prev_col] = next_col # map e.g., query_0 to be renamed to query_1
            prev_col = next_col
        else:
             break
    df = df.rename(columns=rename_cols)
    if keep_original:
        df['query'] = df["query_0"]
    return df


def push_queries_dict(inp: IterDictRecord, *, keep_original: bool = False, inplace: bool = False) -> IterDictRecord:
    """
    Works like ``push_queries`` but over a dict instead of a dataframe.
    """
    if "query" not in inp:
        raise KeyError(f"Expected a query column, but found {list(inp.keys())}")
    if not inplace:
        inp = inp.copy()
    prev_col = 'query'
    for query_idx in itertools.count():
        next_col = f'query_{query_idx}'
        if prev_col in inp:
            inp[next_col] = inp.pop(prev_col) # assign e.g., query_1 to query_0
            prev_col = next_col
        else:
             break
    if keep_original:
        inp['query'] = inp['query_0']
    return inp


def pop_queries(df: pd.DataFrame, *, inplace: bool = False) -> pd.DataFrame:
    """
        Changes a dataframe such that the "query_0" column becomes "query_1", and any
        "query_1" columns becames "query_0" etc. In effect, does the opposite of push_queries().
        The current "query" column is dropped.

        Arguments:
            df: Dataframe with a "query" column
            inplace: if False, a copy of the dataframe is returned. If True, changes are made to the
                supplied dataframe. Defaults to False. 
    """
    cols = set(df.columns)
    if "query_0" not in cols:
        raise KeyError(f"Expected a query_0 column, but found {list(cols)}")
    if not inplace:
        df = df.copy()
    df.drop(columns=["query"], inplace=True)
    prev_col = 'query'
    rename_cols = {}
    for query_idx in itertools.count():
        next_col = f'query_{query_idx}'
        if next_col in cols:
            rename_cols[next_col] = prev_col # map e.g., query_1 to be renamed to query_0
            prev_col = next_col
        else:
             break
    df = df.rename(columns=rename_cols)
    return df


def ranked_documents_to_queries(topics_and_res : pd.DataFrame):
    return topics_and_res[query_columns(topics_and_res, qid=True)].groupby(["qid"]).first().reset_index()


def coerce_queries_dataframe(query):
    """
    Convert either a string or a list of strings to a dataframe for use as topics in retrieval.

    Args:
        query: Either a string or a list of strings

    Returns:
        dataframe with columns=['qid','query']
    """
    if isinstance(query, pd.DataFrame):
        return query
    elif isinstance(query, str):
        return pd.DataFrame([["1", query]], columns=['qid', 'query'])
    # if queries is a list or tuple
    elif isinstance(query, list) or isinstance(query, tuple):
        # if the list or tuple is made of strings
        if query != [] and isinstance(query[0], str):
            indexed_query = []
            for i, item in enumerate(query):
                # all elements must be of same type
                assert isinstance(item, str), f"{item} is not a string"
                indexed_query.append([str(i + 1), item])
            return pd.DataFrame(indexed_query, columns=['qid', 'query'])
    # catch-all when we dont recognise the type
    raise ValueError("Could not coerce %s (type %s) into a DataFrame of queries" % (str(query), str(type(query))))


def coerce_dataframe_types(dataframe):
    """
    Changes data types to match standard values. The dataframe need not have all the columns,
    but if they are present, will cast the values to the proper types.
     - ``qid`` -> ``str``
     - ``docno`` -> ``str``
     - ``score`` -> ``float``

    Args:
        dataframe: a Pandas dataframe

    Returns:
        dataframe with data types properly set
    """
    TYPE_MAP = { # python type -> acceptable numpy types
        str: (np.dtype('O'),),
        float: (np.dtype('float32'), np.dtype('float64')),
    }
    COLUMN_MAP = { # column name -> python type
        'qid': str,
        'docno': str,
        'score': float,
    }
    for column, dtype in COLUMN_MAP.items():
        if column in dataframe.columns and dataframe[column].dtype not in TYPE_MAP[dtype]:
            dataframe[column] = dataframe[column].astype(dtype)
    return dataframe


def split_df(df : pd.DataFrame, N: Optional[int] = None, *, batch_size: Optional[int] = None) -> List[pd.DataFrame]:
    """
    Splits a dataframe into N different chunks. Splitting will be sensitive to the primary datatype
    of the dataframe (Q,R,D).

    Either ``N`` (the number of chunks) or ``batch_size`` (the size of each chunk) should be provided (but not both).
    """
    assert (N is None) != (batch_size is None), "Either N or batch_size should be provided (and not both)"

    if N is None:
        assert batch_size is not None
        N = math.ceil(len(df) / batch_size)

    type = None
    if "qid" in df.columns:
        if "docno" in df.columns:
            type = "R"
        else:
            type = "Q"
    elif "docno" in df.columns:
        type = "D"
    else:
        raise ValueError("Dataframe is not of type D,Q,R")

    def chunks(df, n):
        """Yield successive n-sized chunks from df."""
        for i in range(0, len(df), n):
            yield df.iloc[ i: min(len(df),i + n)]
    
    if type == "Q" or type == "D":         
        splits = list( chunks(df, math.ceil(len(df)/N)))
        return splits

    rtr = []
    grouper = df.groupby("qid")
    this_group = []
    chunk_size = math.ceil(len(grouper)/N)
    for qid, group in grouper:
        this_group.append(group)
        if len(this_group) == chunk_size:
            rtr.append(pd.concat(this_group))
            this_group = []
    if len(this_group) > 0:
        rtr.append(pd.concat(this_group))
    return rtr


_ir_measures_to_pyterrier = {
    'query_id': 'qid',
    'doc_id': 'docno',
    'relevance': 'label',
}

@overload
def from_ir_measures(inp: str) -> str: ...
@overload
def from_ir_measures(inp: Dict[str, Any]) -> Dict[str, Any]: ...
@overload
def from_ir_measures(inp: List[str]) -> List[str]: ...
@overload
def from_ir_measures(inp: Set[str]) -> List[str]: ...
@overload
def from_ir_measures(inp: pd.DataFrame) -> pd.DataFrame: ...
def from_ir_measures(
    inp: Union[str, pd.DataFrame, Dict[str, Any], List[str], Set[str]],
) -> Union[str, pd.DataFrame, Dict[str, Any], List[str]]:
    """This function maps ir-measues column names to PyTerrier column names.

    It's useful when converting between PyTerrier and ir-measures data formats.

    .. seealso::
        :py:func:`pyterrier.model.to_ir_measures` for the reverse operation.
    """
    if isinstance(inp, str):
        return _ir_measures_to_pyterrier.get(inp, inp) # rename values in mapping, keep others the same
    elif isinstance(inp, pd.DataFrame):
        return inp.rename(columns=_ir_measures_to_pyterrier)
    elif isinstance(inp, dict):
        return { _ir_measures_to_pyterrier.get(k, k): v for k, v in inp.items() }
    else:
        return [_ir_measures_to_pyterrier.get(x, x) for x in inp]


_pyterrier_to_ir_measures = {
    'qid': 'query_id',
    'docno': 'doc_id',
    'label': 'relevance',
}

@overload
def to_ir_measures(inp: str) -> str: ...
@overload
def to_ir_measures(inp: Dict[str, Any]) -> Dict[str, Any]: ...
@overload
def to_ir_measures(inp: List[str]) -> List[str]: ...
@overload
def to_ir_measures(inp: pd.DataFrame) -> pd.DataFrame: ...
def to_ir_measures(
    inp: Union[str, pd.DataFrame, Dict[str, Any], List[str]],
) -> Union[str, pd.DataFrame, Dict[str, Any], List[str]]:
    """This function maps PyTerrier column names to ir-measures column names.

    It's useful when converting between PyTerrier and ir-measures data formats.

    .. seealso::
        :py:func:`pyterrier.model.from_ir_measures` for the reverse operation.
    """
    if isinstance(inp, str):
        return _pyterrier_to_ir_measures.get(inp, inp) # rename values in mapping, keep others the same
    elif isinstance(inp, pd.DataFrame):
        return inp.rename(columns=_pyterrier_to_ir_measures)
    elif isinstance(inp, dict):
        return { _pyterrier_to_ir_measures.get(k, k): v for k, v in inp.items() }
    return [_pyterrier_to_ir_measures.get(x, x) for x in inp]


def frame_info(columns : List[str]) -> Optional[Dict[str, str]]:
    """Returns a dict containing a short label and a short description for given set of columns."""
    if 'qid' in columns and 'docno' in columns and 'features' in columns:
        return {
            "label": 'R_f',
            "title": 'Result Frame with Features',
        }
    elif 'qid' in columns and 'docno' in columns:
        return {
            "label": 'R',
            "title": 'Result Frame',
        }
    elif 'qanswer' in columns:
        return {
            "label": 'A',
            "title": 'Query Answer Frame',
        }
    elif 'qid' in columns:
        return {
            "label": 'Q',
            "title": 'Query Frame',
        }
    elif 'docno' in columns:
        return {
            "label": 'D',
            "title": 'Document Frame',
        }
    return None

def column_info(column: str) -> Optional[dict]:
    """Returns a dictionary with information about the specified column name."""
    if column == 'qid':
        return {
            'title': 'qid',
            'phrase': 'Query ID',
            'short_desc': 'ID of query in frame',
            'type': str,
        }
    if column == 'docno':
        return {
            'title': 'docno',
            'phrase': 'External Document ID',
            'short_desc': 'String ID of document in collection',
            'type': str,
        }
    if column == 'docid':
        return {
            'title': 'docid',
            'phrase': 'Internal Document ID',
            'short_desc': 'Integer ID of document in a specific index',
            'type': int,
        }
    if column == 'score':
        return {
            'title': 'score',
            'short_desc': 'Ranking score of document to query (higher=better)',
            'type': float,
        }
    if column == 'rank':
        return {
            'title': 'rank',
            'short_desc': 'Ranking order of document to query (lower=better)',
            'type': int,
        }
    if column == 'query':
        return {
            'title': 'query',
            'short_desc': 'Query text',
            'type': str,
        }
    if re.match(r'^query_[0-9]+$', column):
        return {
            'title': str(column),
            'short_desc': 'Stashed query text',
            'type': str,
        }
    if column == 'text':
        return {
            'title': "text",
            'short_desc': 'Document text',
            'type': str,
        }
    if column == 'title':
        return {
            'title': "title",
            'short_desc': 'Document title',
            'type': str,
        }
    if column == 'qanswer':
        return {
            'title': "qanswer",
            'short_desc': 'Answer to the query',
            'type': str,
        }
    if column == 'qcontext':
        return {
            'title': "qcontext",
            'short_desc': 'Context to the query',
            'type': str,
        }
    if column == 'features':
        return {
            'title': "features",
            'short_desc': 'Feature array for learning-to-rank',
            'type': np.array,
        }
    if column == 'query_vec':
        return {
            'title': "query_vec",
            'short_desc': 'Dense query vector',
            'type': np.array,
        }
    if column == 'doc_vec':
        return {
            'title': "doc_vec",
            'short_desc': 'Dense document vector',
            'type': np.array,
        }
    if column == 'query_toks':
        return {
            'title': "query_toks",
            'short_desc': 'Sparse query vector',
            'type': dict,
        }
    if column == 'toks':
        return {
            'title': "toks",
            'short_desc': 'Sparse document vector',
            'type': dict,
        }
    return None



import pyterrier as pt
import os
import unittest
from .base import BaseTestCase

class TestLTRPipeline(BaseTestCase):

    def test_fastrank(self):
        try:
            import fastrank
        except:
            self.skipTest("Fastrank not installed")
        train_request = fastrank.TrainRequest.coordinate_ascent()
        params = train_request.params
        params.init_random = True
        params.normalize = True
        params.seed = 1234567

        topics = pt.io.read_topics(self.here + "/fixtures/vaswani_npl/query_light.trec").head(5)
        qrels = pt.io.read_qrels(self.here + "/fixtures/vaswani_npl/qrels")

        pipeline = pt.terrier.FeaturesRetriever(self.here + "/fixtures/index/data.properties", ["WMODEL:PL2", "WMODEL:BM25"], controls={"wmodel" : "DPH"}) >> \
            pt.ltr.apply_learned_model(train_request, form="fastrank")
        
        pipeline.fit(topics, qrels, topics, qrels)
        pt.Evaluate(
            pipeline.transform(topics),
            qrels
        )

    def test_xgltr_pipeline(self):
        try:
            import xgboost as xgb
        except:
            self.skipTest("xgboost not installed")

        xgparams = {
            'objective': 'rank:ndcg',
            'learning_rate': 0.1,
            'gamma': 1.0, 'min_child_weight': 0.1,
            'max_depth': 6,
            'random_state': 42
        }

        topics = pt.io.read_topics(self.here + "/fixtures/vaswani_npl/query_light.trec").head(5)
        qrels = pt.io.read_qrels(self.here + "/fixtures/vaswani_npl/qrels")

        features = pt.terrier.FeaturesRetriever(self.here + "/fixtures/index/data.properties", ["WMODEL:PL2", "WMODEL:BM25"], controls={"wmodel" : "DPH"})
        pipeline = features >> pt.ltr.apply_learned_model(xgb.sklearn.XGBRanker(**xgparams), form="ltr")
        
        pipeline.fit(topics, qrels, topics, qrels)
        pt.Evaluate(
            pipeline.transform(topics),
            qrels
        )
  
    def test_ltr_mocked(self):

        class MockLTR():
            def fit(self, 
                    features_tr, labels_tr, 
                    group=None, eval_set=None, eval_group=None):
                self.features_tr = features_tr
                self.labels_tr = labels_tr
                self.group = group
                self.eval_set = eval_set
                self.eval_group = eval_group
        
        import pandas as pd, numpy as np
        res = pd.DataFrame([ 
                ["q2", "b", np.array([2])], # feature value encodes the number of docs for this query
                ["q2", "a", np.array([2])], 
                ["q1", "a", np.array([1])] 
            ], columns=["qid", "docno", "features"])
        qrels = pd.DataFrame([ ["q2", "a", 1], ["q1", "a", 1] ], columns=["qid", "docno", "label"])

        mockltr = MockLTR()
        mock_pipeline = pt.ltr.apply_learned_model(mockltr, form="ltr")
        mock_pipeline.fit(res, qrels, res, qrels)

        # check feature shapes
        self.assertEqual(3, mockltr.features_tr.shape[0])
        self.assertEqual(1, mockltr.features_tr.shape[1])
        # check label shape
        self.assertEqual(3, mockltr.labels_tr.shape[0])
        # check label counts
        self.assertEqual(2, np.count_nonzero(mockltr.labels_tr == 1))
        self.assertEqual(1, np.count_nonzero(mockltr.labels_tr == 0))

        # idea here is to check ordering between group and features information
        for i in range(2): # 2 queries
            self.assertEqual(mockltr.features_tr[i, 0], mockltr.group[i], "at query index %d" % i)

    def test_skltr_pipeline(self):
        from sklearn.ensemble import RandomForestClassifier

        topics = pt.io.read_topics(self.here + "/fixtures/vaswani_npl/query_light.trec").head(5)
        qrels = pt.io.read_qrels(self.here + "/fixtures/vaswani_npl/qrels")

        pipeline = ( pt.terrier.FeaturesRetriever(self.here + "/fixtures/index/data.properties", ["WMODEL:PL2", "WMODEL:BM25"], controls={"wmodel" : "DPH"}) >>
            pt.ltr.apply_learned_model(RandomForestClassifier())
        )
        
        pipeline.fit(topics, qrels)
        pt.Evaluate(
            pipeline.transform(topics),
            qrels
        )

    def test_ltr_pipeline_feature_change(self):
        from sklearn.ensemble import RandomForestClassifier

        topics = pt.io.read_topics(self.here + "/fixtures/vaswani_npl/query_light.trec").head(5)
        qrels = pt.io.read_qrels(self.here + "/fixtures/vaswani_npl/qrels")

        rf = RandomForestClassifier()

        pipeline = pt.terrier.FeaturesRetriever(self.here + "/fixtures/index/data.properties", ["WMODEL:PL2", "WMODEL:BM25"], controls={"wmodel" : "DPH"}) >> \
            pt.ltr.apply_learned_model(rf)
        
        pipeline.fit(topics, qrels)
        pipeline.transform(topics)

        pipeline2 = pt.terrier.FeaturesRetriever(self.here + "/fixtures/index/data.properties", ["WMODEL:PL2", "WMODEL:BM25", "WMODEL:Dl"], controls={"wmodel" : "DPH"}) >> \
            pt.ltr.apply_learned_model(rf)
        with self.assertRaises(ValueError):
            pipeline2.transform(topics)

if __name__ == "__main__":
    unittest.main()



from pyterrier import Transformer
from typing import Dict, List, Literal, Optional, Tuple, Union, overload
from ._execution import _run_and_evaluate
from ._utils import _restore_state, _save_state
from . import MEASURE_TYPE, MEASURES_TYPE
from ir_measures import Measure
import pandas as pd
import pyterrier as pt

TRANSFORMER_PARAMETER_VALUE_TYPE = Union[str,float,int,str]
GRID_SCAN_PARAM_SETTING = Tuple[
            Transformer, 
            str, 
            TRANSFORMER_PARAMETER_VALUE_TYPE
        ]
GRID_SEARCH_RETURN_TYPE_SETTING = Tuple[
    float, 
    List[GRID_SCAN_PARAM_SETTING]
]

GRID_SEARCH_RETURN_TYPE_BOTH = Tuple[
    Transformer,
    float, 
    List[GRID_SCAN_PARAM_SETTING]
]


@overload
def GridSearch(
        pipeline : Transformer,
        params : Dict[Transformer,Dict[str,List[TRANSFORMER_PARAMETER_VALUE_TYPE]]],
        topics : pd.DataFrame,
        qrels : pd.DataFrame,
        metric : MEASURE_TYPE,
        jobs : int,
        backend: str,
        verbose: bool ,
        batch_size : Optional[int],
        return_type : Literal['opt_pipeline'],
    ) -> Transformer: ...

@overload
def GridSearch(
        pipeline : Transformer,
        params : Dict[Transformer,Dict[str,List[TRANSFORMER_PARAMETER_VALUE_TYPE]]],
        topics : pd.DataFrame,
        qrels : pd.DataFrame,
        metric : MEASURE_TYPE,
        jobs : int,
        backend: str,
        verbose: bool ,
        batch_size : Optional[int],
        return_type : Literal['best_setting'],
    ) -> GRID_SEARCH_RETURN_TYPE_SETTING: ...

@overload
def GridSearch(
        pipeline : Transformer,
        params : Dict[Transformer,Dict[str,List[TRANSFORMER_PARAMETER_VALUE_TYPE]]],
        topics : pd.DataFrame,
        qrels : pd.DataFrame,
        metric : MEASURE_TYPE,
        jobs : int,
        backend: str,
        verbose: bool ,
        batch_size : Optional[int],
        return_type : Literal['both'],
    ) -> GRID_SEARCH_RETURN_TYPE_BOTH: ...

def GridSearch(
        pipeline : Transformer,
        params : Dict[Transformer,Dict[str,List[TRANSFORMER_PARAMETER_VALUE_TYPE]]],
        topics : pd.DataFrame,
        qrels : pd.DataFrame,
        metric : MEASURE_TYPE = "map",
        jobs : int = 1,
        backend='joblib',
        verbose: bool = False,
        batch_size : Optional[int] = None,
        return_type : Literal['opt_pipeline', 'best_setting', 'both'] = "opt_pipeline"
    ) -> Union[Transformer,GRID_SEARCH_RETURN_TYPE_SETTING,GRID_SEARCH_RETURN_TYPE_BOTH]:
    """
    GridSearch is essentially, an argmax GridScan(), i.e. it returns an instance of the pipeline to tune
    with the best parameter settings among params, that were found that were obtained using the specified
    topics and qrels, and for the specified measure.

    :param pipeline: a transformer or pipeline to tune
    :param params: a two-level dictionary, mapping transformer to param name to a list of values
    :param topics: topics to tune upon
    :param qrels: qrels to tune upon       
    :param metric: name of the metric on which to determine the most effective setting. Defaults to "map".
    :param batch_size: If not None, evaluation is conducted in batches of batch_size topics. Default=None, which evaluates all topics at once. 
        Applying a batch_size is useful if you have large numbers of topics, and/or if your pipeline requires large amounts of temporary memory
        during a run. Default is None.
    :param jobs: Number of parallel jobs to run. Default is 1, which means sequentially.
    :param backend: Parallelisation backend to use. Defaults to "joblib". 
    :param verbose: whether to display progress bars or not
    :param return_type: whether to return the same transformer with optimal pipeline setting, and/or a setting of the
        higher metric value, and the resulting transformers and settings.
    """
    # save state
    initial_state = _save_state(params)

    if isinstance(metric, list):
        raise KeyError("GridSearch can only maximise ONE metric, but you passed a list (%s)." % str(metric))

    grid_outcomes = GridScan(
        pipeline, 
        params, 
        topics, 
        qrels, 
        [metric], 
        jobs, 
        backend, 
        verbose, 
        batch_size, 
        dataframe=False)
    assert not isinstance(grid_outcomes, pd.DataFrame)

    assert len(grid_outcomes) > 0, "GridScan returned 0 rows"
    max_measure = grid_outcomes[0][1][metric]
    max_setting = grid_outcomes[0][0]
    for setting, measures in grid_outcomes: # TODO what is the type of this iteration?
        if measures[metric] > max_measure:
            max_measure = measures[metric]
            max_setting = setting
    print("Best %s is %f" % (metric, max_measure))
    print("Best setting is %s" % str(["%s %s=%s" % (str(t), k, v) for t, k, v in max_setting]))

    if return_type == "opt_pipeline":
        for tran, param, value in max_setting:
            tran.set_parameter(param, value)
        return pipeline
    if return_type == "best_setting":
        _restore_state(initial_state)
        return max_measure, max_setting
    if return_type == "both":
        for tran, param, value in max_setting:
            tran.set_parameter(param, value)
        return (pipeline, max_measure, max_setting)
    raise ValueError("Unknown return_type option %s" % return_type)

def GridScan(
        pipeline : Transformer,
        params : Dict[Transformer,Dict[str,List[TRANSFORMER_PARAMETER_VALUE_TYPE]]],
        topics : pd.DataFrame,
        qrels : pd.DataFrame,
        metrics : Union[MEASURE_TYPE,MEASURES_TYPE] = ["map"],
        jobs : int = 1,
        backend='joblib',
        verbose: bool = False,
        batch_size = None,
        dataframe = True,
    ) -> Union[pd.DataFrame, List [ Tuple [ List[ GRID_SCAN_PARAM_SETTING ], Dict[Union[str, Measure] ,float]  ]  ] ]:
    """
    GridScan applies a set of named parameters on a given pipeline and evaluates the outcome. The topics and qrels 
    must be specified. The trec_eval measure names can be optionally specified.
    The transformers being tuned, and their respective parameters are named in the param_dict. The parameter being
    varied must be changable using the :func:`set_parameter()` method. This means instance variables,
    as well as controls in the case of Retriever.

    :param pipeline: a transformer or pipeline
    :param params: a two-level dictionary, mapping transformer to param name to a list of values
    :param topics: topics to tune upon
    :param qrels: qrels to tune upon       
    :param metrics): name of the metrics to report for each setting. Defaults to ["map"].
    :param batch_size: If not None, evaluation is conducted in batches of batch_size topics. Default=None, which evaluates all topics at once. 
        Applying a batch_size is useful if you have large numbers of topics, and/or if your pipeline requires large amounts of temporary memory
        during a run. Default is None.
    :param jobs: Number of parallel jobs to run. Default is 1, which means sequentially.
    :param backend: Parallelisation backend to use. Defaults to "joblib". 
    :param verbose: whether to display progress bars or not
    :param dataframe: return a dataframe or a list

    :return: A dataframe showing the effectiveness of all evaluated settings, if dataframe=True
        A list of settings and resulting evaluation measures, if dataframe=False
    
    Raises:
        ValueError: if a specified transformer does not have such a parameter

    Example::

        # graph how PL2's c parameter affects MAP
        pl2 = pt.terrier.Retriever(index, wmodel="PL2", controls={'c' : 1})
        rtr = pt.GridScan(
            pl2, 
            {pl2 : {'c' : [0.1, 1, 5, 10, 20, 100]}}, 
            topics,
            qrels,
            ["map"]
        )
        import matplotlib.pyplot as plt
        plt.plot(rtr["tran_0_c"], rtr["map"])
        plt.xlabel("PL2's c value")
        plt.ylabel("MAP")
        plt.show()

    """
    import itertools

    if verbose and jobs > 1:
        from warnings import warn
        warn("Cannot provide progress on parallel job")
    if isinstance(metrics, str):
        metrics = [metrics]

    # Store the all parameter names and candidate values into a dictionary, keyed by a tuple of the transformer and the parameter name
    # such as {(Retriever, 'wmodel'): ['BM25', 'PL2'], (Retriever, 'c'): [0.1, 0.2, 0.3], (Bla, 'lr'): [0.001, 0.01, 0.1]}
    candi_dict: Dict[Tuple[Transformer, str], List[TRANSFORMER_PARAMETER_VALUE_TYPE]] = {}
    for tran, param_set in params.items():
        for param_name, values in param_set.items():
            candi_dict[ (tran, param_name) ] = values
    if len(candi_dict) == 0:
        raise ValueError("No parameters specified to optimise")
    for tran, param in candi_dict:
        try:
            tran.get_parameter(param)
        except Exception:
            raise ValueError("Transformer %s does not expose a parameter named %s" % (str(tran), param))
    
    keys, vals = zip(*candi_dict.items())
    combinations = list(itertools.product(*vals))
    assert len(combinations) > 0, "No combinations selected"

    def _evaluate_one_setting(keys, values):
        #'params' is every combination of candidates
        params = dict(zip(keys, values))
        parameter_list = []
        
        # Set the parameter value in the corresponding transformer of the pipeline
        for (tran, param_name), value in params.items():
            tran.set_parameter(param_name, value)
            # such as (Retriever, 'wmodel', 'BM25')
            parameter_list.append( (tran, param_name, value) )
            
        time, eval_scores = _run_and_evaluate(pipeline, topics, qrels, metrics, perquery=False, batch_size=batch_size)
        return parameter_list, eval_scores

    def _evaluate_several_settings(inputs : List[Tuple]):
        return [_evaluate_one_setting(k,v) for k, v in inputs]

    eval_list = []
    #for each combination of parameter values
    if jobs == 1:
        for v in pt.tqdm(combinations, total=len(combinations), desc="GridScan", mininterval=0.3) if verbose else combinations:
            parameter_list, eval_scores = _evaluate_one_setting(keys, v)
            eval_list.append( (parameter_list, eval_scores) )
    else:
        import itertools
        import more_itertools
        try:
            from pyterrier_alpha.parallel import parallel_lambda # type: ignore
        except ImportError as ie:
            raise ImportError("pyterrier-alpha[parallel] must be installed for jobs>1") from ie
    
        all_inputs = [(keys, values) for values in combinations]

        # how many jobs to distribute this to
        num_batches = int(len(combinations)/jobs) if len(combinations) >= jobs else len(combinations)

        # built the batches to distribute
        batched_inputs = list(more_itertools.chunked(all_inputs, num_batches))
        assert len(batched_inputs) > 0, "No inputs identified for parallel_lambda"
        eval_list = parallel_lambda(_evaluate_several_settings, batched_inputs, jobs, backend=backend)
        eval_list =  list(itertools.chain(*eval_list))
        assert len(eval_list) > 0, "parallel_lambda returned 0 rows" 
    
    # resulting eval_list has the form [ 
    #   ( [(BR, 'wmodel', 'BM25'), (BR, 'c', 0.2)]  ,   {"map" : 0.2654} )
    # ]
    # ie, a list of possible settings, combined with measure values    
    if not dataframe:
        return eval_list

    rtr=[]
    for setting, measures in eval_list:
        row={}
        for i, (tran, param, value) in enumerate(setting):
            row["tran_%d" % i]  = tran
            row["tran_%d_%s" % (i,param) ]  = value
        row.update(measures)
        rtr.append(row)
    # resulting dataframe looks like:
    #    tran_0  tran_0_c       map
    #0  BR(PL2)     0.1  0.104820
    #1  BR(PL2)     1.0  0.189274
    #2  BR(PL2)     5.0  0.230838
    return pd.DataFrame(rtr)

def KFoldGridSearch(
        pipeline : Transformer,
        params : Dict[Transformer,Dict[str,List[TRANSFORMER_PARAMETER_VALUE_TYPE]]],
        topics_list : List[pd.DataFrame],
        qrels : Union[pd.DataFrame,List[pd.DataFrame]],
        metric : MEASURE_TYPE = "map",
        jobs : int = 1,
        backend='joblib',
        verbose: bool = False,
        batch_size : Optional[int] = None) -> Tuple[pd.DataFrame, List[List[GRID_SCAN_PARAM_SETTING]]]:
    """
    Applies a GridSearch using different folds. It returns the *results* of the 
    tuned transformer pipeline on the test topics. The number of topics dataframes passed
    to topics_list defines the number of folds. For each fold, all but one of the dataframes
    is used as training, and the remainder used for testing. 

    The state of the transformers in the pipeline is restored after the KFoldGridSearch has
    been executed.

    :param pipeline: a transformer or pipeline to tune
    :param params: a two-level dictionary, mapping transformer to param name to a list of values
    :param topics_list: a *list* of topics dataframes to tune upon
    :param qrels: qrels to tune upon. A single dataframe, or a list for each fold.       
    :param metric: name of the metric on which to determine the most effective setting. Defaults to "map".
    :param batch_size: If not None, evaluation is conducted in batches of batch_size topics. Default=None, which evaluates all topics at once. 
        Applying a batch_size is useful if you have large numbers of topics, and/or if your pipeline requires large amounts of temporary memory
        during a run. Default is None.
    :param jobs: Number of parallel jobs to run. Default is 1, which means sequentially.
    :param backend: Parallelisation backend to use. Defaults to "joblib". 
    :param verbose(bool): whether to display progress bars or not

    :return: A tuple containing, firstly, the results of pipeline on the test topics after tuning, and secondly, a list of the best parameter settings for each fold.

    Consider tuning a terrier.Retriever PL2 where the folds of queries are pre-determined::

        pl2 = pt.terrier.Retriever(index, wmodel="PL2", controls={'c' : 1})
        tuned_pl2, _ = pt.KFoldGridSearch(
            pl2, 
            {pl2 : {'c' : [0.1, 1, 5, 10, 20, 100]}}, 
            [topicsf1, topicsf2],
            qrels,
            ["map"]
        )
        pt.Experiment([pl2, tuned_pl2], all_topics, qrels, ["map"])

    As 2 splits are defined, PL2 is first tuned on topicsf1 and tested on topicsf2, then 
    trained on topicsf2 and tested on topicsf1. The results dataframe of PL2 after tuning of the c
    parameter are returned by the KFoldGridSearch, and can be used directly in a pt.Experiment().
    """

    import pandas as pd
    num_folds = len(topics_list)
    if isinstance(qrels, pd.DataFrame):
        qrels = [qrels] * num_folds

    FOLDS=list(range(0, num_folds))
    results : List[pd.DataFrame] = []
    settings=[]

    # save state
    initial_state = _save_state(params)

    for fold in FOLDS:
        print("Fold %d" % (fold+1))

        train_indx = FOLDS.copy()
        train_indx.remove(fold)
        train_topics = pd.concat([topics_list[offset] for offset in train_indx])
        train_qrels = pd.concat([qrels[offset] for offset in train_indx])
        test_topics = topics_list[fold]
        #test_qrels arent needed
        #test_qrels = qrels[fold]

        # safety - give the GridSearch a stable initial setting
        _restore_state(initial_state)

        optPipe: Transformer
        max_measure: float
        max_setting: List[GRID_SCAN_PARAM_SETTING]
        optPipe, max_measure, max_setting = GridSearch(
            pipeline,
            params,
            train_topics,
            train_qrels,
            metric,
            jobs=jobs,
            backend=backend,
            verbose=verbose,
            batch_size=batch_size,
            return_type="both")
        results.append(optPipe.transform(test_topics))
        settings.append(max_setting)

    # restore state
    _restore_state(initial_state)

    return (pd.concat(results), settings)


import pyterrier as pt
from pyterrier._evaluation._rendering import _convert_measures
from . import MEASURES_TYPE, SYSTEM_OR_RESULTS_TYPE, VALIDATE_TYPE


import ir_measures
import pandas as pd


from typing import Sequence
from warnings import warn


def _validate(
        retr_systems : Sequence[SYSTEM_OR_RESULTS_TYPE],
        topics : pd.DataFrame,
        eval_metrics : MEASURES_TYPE,
        names : Sequence[str],
        validate : VALIDATE_TYPE):

    if validate == 'ignore':
        return

    # convert metrics to a list of Measure objects
    # NB: this mapping is already performed in _run_and_evaluate, but we need it here to 
    # validate the transformers; I think its inexpensive
    _metrics, rev_mapping = _convert_measures(eval_metrics)

    required_cols = set(ir_measures.run_inputs(_metrics)) # find required columns for the measures (usually: query_id, doc_id, score)
    required_cols = set(list(pt.model.from_ir_measures(required_cols))) # convert to pyterrier naming conventions (usually: qid, docno, score)

    for i, (name, system) in enumerate(zip(names, retr_systems)):
        friendly_name = "Transformer %s (%s) at position %i" % (name, str(system), i) if name != str(system) else "Transformer %s at position %i" % (str(system), i)
        found_cols = []
        if isinstance(system, pd.DataFrame):
            found_cols = system.columns.tolist()
        elif isinstance(system, pt.Transformer):
            try:
                found_cols = pt.inspect.transformer_outputs(system, input_columns=topics.columns.tolist()) or [] # or [] will never be called, but is needed to make mypy happy
            except pt.inspect.InspectError as ie: # when we cant tell
                warn(
                    "%s failed to validate: %s - if your pipeline works, set validate='ignore' to remove this warning, or add transform_output method to the transformers in this pipeline to clarify how it works" % (friendly_name, str(ie)))
                continue
            except pt.validate.InputValidationError as ie: # (when input validation fails)
                if validate == 'warn':
                    warn(
                        "%s failed to validate: %s" % (friendly_name, str(ie)))
                    continue
                elif validate == 'error':
                    raise ValueError("%s failed to validate: %s" % (friendly_name, str(ie))) from ie
        else:
            raise TypeError("Expected a list of Transformers or DataFrames, but received unexpected type %s for retrieval system at position %d" % (str(type(system)), i))

        if required_cols.difference(found_cols):
            message = "Transformer %s (%s) at position %i does not produce all required columns %s, found only %s" % (
                name, str(system), i, str(list(required_cols)), str(found_cols))
            if validate == 'warn':
                warn(
                    message)
            elif validate == 'error':
                raise ValueError(message)


import unittest
import pyterrier as pt, pandas as pd
from .base import BaseTestCase

class TestDebug(BaseTestCase):

    def test_num_rows(self):
        df = pt.new.queries(["a"])
        import io
        import contextlib
        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            pt.debug.print_num_rows().transform(df)
        self.assertTrue("num_rows 1: 1" in f.getvalue())


import os
import sys
from pyterrier.java import required_raise, required, before_init, started, mavenresolver, JavaClasses, JavaInitializer, register_config
from typing import Optional
import pyterrier as pt

_min_colab_jdk = "openjdk-11-jdk-headless"
_stdout_ref = None
_stderr_ref = None


# ----------------------------------------------------------
# Java Initialization
# ----------------------------------------------------------



class ColabJavaInit(JavaInitializer):
    def priority(self) -> int:
        return -101 # run this initializer before CoreJavaInit
    
    def pre_init(self, jnius_config):
        import sys
        # detect colab
        if 'google.colab' not in sys.modules:
            return
        import shutil
        # detect java on the PATH
        if shutil.which("java") is not None:
            return
        print(f"This Colab is missing Java - installing {_min_colab_jdk}, please wait")
        import subprocess
        import os

        cmd = [
            "apt-get", 
            "install", 
            "-y", 
            _min_colab_jdk,
            "--option=Dpkg::Progress-Fancy=1",
            "--option=APT::Color=1"
        ]

        process = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
            bufsize=1,
            env={**os.environ, "TERM": "xterm-color"}
        )

        for line in process.stdout:
            sys.stdout.write(line)
            sys.stdout.flush()

        process.wait()
        # ✅ Check exit status
        if process.returncode == 0:
            print(f"\n✅ apt-get install of {_min_colab_jdk} completed successfully.")
        else:
            print(f"\n❌ apt-get install of {_min_colab_jdk} failed with exit code {process.returncode}.")

class CoreJavaInit(JavaInitializer):
    def priority(self) -> int:
        return -100 # run this initializer before anything else

    def pre_init(self, jnius_config):
        if configure['java_home']:
            os.environ['JAVA_HOME'] = configure['java_home']

        if pt.utils.is_windows():
            if "JAVA_HOME" in os.environ:
                java_home =  os.environ["JAVA_HOME"]
                fix = f'{java_home}\\jre\\bin\\server\\;{java_home}\\jre\\bin\\client\\;{java_home}\\bin\\server\\'
                os.environ["PATH"] = os.environ["PATH"] + ";" + fix

        if pt.java.configure['mem'] is not None:
            jnius_config.add_options('-Xmx' + str(pt.java.configure['mem']) + 'm')

        for opt in pt.java.configure['options']:
            jnius_config.add_options(opt)

        for jar in pt.java.configure['jars']:
            jnius_config.add_classpath(jar)

        # set the property that makes a process name visible in jps
        process_name : str =  pt.utils._get_notebook()
        if process_name is None:
            process_name = "python[pyterrier]:" + (sys.argv[0] if sys.argv[0] else '<interactive>')
        else:
            process_name = "jupyter[pyterrier]:" + process_name
        jnius_config.add_options("-Dsun.java.command=%s" % process_name)

    @required_raise
    def post_init(self, jnius):
        pt.java.set_log_level(pt.java.configure['log_level'])

        if pt.java.configure['redirect_io']:
            pt.java.redirect_stdouterr()

        java_version = pt.java.J.System.getProperty("java.version")
        if java_version.startswith("1.") or java_version.startswith("9."):
            raise RuntimeError(f"Pyterrier requires Java 11 or newer, we only found Java version {java_version};"
                + " install a more recent Java, or change os.environ['JAVA_HOME'] to point to the proper Java installation")

        jnius.protocol_map['java.util.Map$Entry'] = {
            '__getitem__' : _mapentry_getitem,
            '__iter__' : lambda self: iter([self.getKey(), self.getValue()]),
            '__len__' : lambda self: 2
        }


# Map$Entry can be decoded like a tuple
def _mapentry_getitem(self, i):
    if i == 0:
        return self.getKey()
    if i == 1:
        return self.getValue()
    raise IndexError()

class Java24Init(JavaInitializer):
    """Responsible for hacking around JDK safety checks from JDK 24 onwards"""
    def priority(self) -> int:
        return -99 # run this before TerrierJavaInit
    
    def pre_init(self, jnius_config):
        # detect JDK 24 onwards - this is a best attempt - at best, the user will see a warning.
        # the plan is to use https://github.com/kivy/pyjnius/pull/780 to ask the JVM for its version
        # before starting the JVM. 
        # We /could/ safely add this option from at least JDK 21 onwards. JDK 11 doesnt recognise it.
        if "JAVA_HOME" in os.environ:
            java_home = os.environ["JAVA_HOME"]
            if any([f"{ver}." in java_home for ver in [24,25,26,28,29]]):
                jnius_config.add_options("--enable-native-access=ALL-UNNAMED")

    @required_raise
    def post_init(self, jnius):
        from packaging.version import Version, parse
        import re
        java_version = pt.java.J.System.getProperty("java.version")
        # RTD has an annoying -internal in their Java version
        java_version = re.sub(r'[-_].*$', '', java_version)
        if parse(java_version) >= Version("24"):
            # Hadoop will fallback to pureJava for sparc architecture - lets pretend we are for just a minute.
            arch = pt.java.J.System.getProperty("os.arch")
            pt.java.J.System.setProperty("os.arch", "sparc")
            # force initialisation of FastByteComparisons using the sparc arch
            pt.java.autoclass("org.apache.hadoop.io.FastByteComparisons$LexicographicalComparerHolder")
            pt.java.J.System.setProperty("os.arch", arch)

def _is_binary(f):
    import io
    return isinstance(f, (io.RawIOBase, io.BufferedIOBase))


@required
def redirect_stdouterr():
    from jnius import autoclass, PythonJavaClass, java_method

    # TODO: encodings may be a probem here
    class MyOut(PythonJavaClass):
        __javainterfaces__ = ['org.terrier.python.OutputStreamable']

        def __init__(self, pystream):
            super(MyOut, self).__init__()
            self.pystream = pystream
            self.binary = _is_binary(pystream)

        @java_method('()V')
        def close(self):
            self.pystream.close()

        @java_method('()V')
        def flush(self):
            self.pystream.flush()

        @java_method('([B)V', name='write')
        def writeByteArray(self, byteArray):
            # TODO probably this could be faster.
            for c in byteArray:
                self.writeChar(c)

        @java_method('([BII)V', name='write')
        def writeByteArrayIntInt(self, byteArray, offset, length):
            # TODO probably this could be faster.
            for i in range(offset, offset + length):
                self.writeChar(byteArray[i])

        @java_method('(I)V', name='write')
        def writeChar(self, chara):
            if self.binary:
                return self.pystream.write(bytes([chara]))
            return self.pystream.write(chr(chara))

    # we need to hold lifetime references to _stdout_ref/_stderr_ref, to ensure
    # they arent GCd. This prevents a crash when Java callsback to  GCd py obj

    global _stdout_ref
    global _stderr_ref
    import sys
    _stdout_ref = MyOut(sys.stdout)
    _stderr_ref = MyOut(sys.stderr)
    jls = autoclass("java.lang.System")
    jls.setOut(
        autoclass('java.io.PrintStream')(
            autoclass('org.terrier.python.ProxyableOutputStream')(_stdout_ref),
            signature="(Ljava/io/OutputStream;)V"))
    jls.setErr(
        autoclass('java.io.PrintStream')(
            autoclass('org.terrier.python.ProxyableOutputStream')(_stderr_ref),
            signature="(Ljava/io/OutputStream;)V"))


def bytebuffer_to_array(buffer):
    assert buffer is not None
    def unsign(signed):
        return signed + 256 if signed < 0 else signed
    return bytearray([ unsign(buffer.get(offset)) for offset in range(buffer.capacity()) ])


# ----------------------------------------------------------
# Configuration
# ----------------------------------------------------------

configure = register_config('pyterrier.java', {
    'jars': [],
    'options': [],
    'mem': None,
    'log_level': 'WARN',
    'redirect_io': True,
    'java_home': None,
})


@before_init
def add_jar(jar_path):
    configure.append('jars', jar_path)


@before_init
def add_package(org_name : str, package_name : str, version : Optional[str] = None, file_type : str = 'jar'):
    if version is None or version == 'snapshot':
        version = mavenresolver.latest_version_num(org_name, package_name)
    file_name = mavenresolver.get_package_jar(org_name, package_name, version, artifact=file_type)
    add_jar(file_name)


@before_init
def set_memory_limit(mem: Optional[float]):
    configure['mem'] = mem


@before_init
def add_option(option: str):
    configure.append('options', option)


@before_init
def set_redirect_io(redirect_io: bool):
    configure['redirect_io'] = redirect_io


@before_init
def set_java_home(java_home: str):
    """
    Sets the directory to search when loading Java.

    Note that you can achieve the same outcome by setting the `JAVA_HOME` environment variable.
    """
    configure['java_home'] = java_home

def set_log_level(level):
    """
        Set the logging level. The following string values are allowed, corresponding
        to Java logging levels:
        
         - `'ERROR'`: only show error messages
         - `'WARN'`: only show warnings and error messages (default)
         - `'INFO'`: show information, warnings and error messages
         - `'DEBUG'`: show debugging, information, warnings and error messages
        
        Unlike other java settings, this can be changed either before or after init() has been called.
    """
    if not started():
        configure['log_level'] = level
    else:
        J.PTUtils.setLogLevel(level, None) # noqa: PT100 handled by started() check above


# ----------------------------------------------------------
# Common classes (accessible via pt.java.J.[ClassName])
# ----------------------------------------------------------

J = JavaClasses(
    ArrayList = 'java.util.ArrayList',
    Properties = 'java.util.Properties',
    PTUtils = 'org.terrier.python.PTUtils',
    System = 'java.lang.System',
    StringReader = 'java.io.StringReader',
    HashMap = 'java.util.HashMap',
    Arrays = 'java.util.Arrays',
    Array = 'java.lang.reflect.Array',
    String = 'java.lang.String',
    List = 'java.util.List',
)
